from typing import Optional, List, Dict
from typing_extensions import Literal

from pydantic import BaseModel

from phidata.infra.k8s.enums.api_version import ApiVersion
from phidata.infra.k8s.enums.kind import Kind
from phidata.infra.k8s.enums.pv import PVAccessMode
from phidata.infra.k8s.enums.volume_type import VolumeType
from phidata.infra.k8s.resource.core.v1.persistent_volume import (
    PersistentVolume,
    PersistentVolumeSpec,
    VolumeNodeAffinity,
    GcePersistentDiskVolumeSource,
    LocalVolumeSource,
    HostPathVolumeSource,
)
from phidata.infra.k8s.create.common.labels import create_component_labels_dict
from phidata.infra.k8s.resource.meta.v1.object_meta import ObjectMeta
from phidata.utils.cli_console import print_error, print_info
from phidata.utils.log import logger


class CreatePersistentVolume(BaseModel):
    pv_name: str
    app_name: str
    namespace: Optional[str] = None
    labels: Optional[Dict[str, str]] = None
    # AccessModes contains all ways the volume can be mounted.
    # More info: https://kubernetes.io/docs/concepts/storage/persistent-volumes#access-modes
    access_modes: List[PVAccessMode] = [PVAccessMode.READ_WRITE_ONCE]
    capacity: Optional[Dict[str, str]] = None
    # A list of mount options, e.g. ["ro", "soft"]. Not validated - mount will simply fail if one is invalid.
    # More info: https://kubernetes.io/docs/concepts/storage/persistent-volumes/#mount-options
    mount_options: Optional[List[str]] = None
    # NodeAffinity defines constraints that limit what nodes this volume can be accessed from.
    # This field influences the scheduling of pods that use this volume.
    node_affinity: Optional[VolumeNodeAffinity] = None
    # What happens to a persistent volume when released from its claim.
    #   The default policy is Retain.
    persistent_volume_reclaim_policy: Optional[
        Literal["Delete", "Recycle", "Retain"]
    ] = None
    # Name of StorageClass to which this persistent volume belongs.
    # Empty value means that this volume does not belong to any StorageClass.
    storage_class_name: Optional[str] = None
    volume_mode: Optional[str] = None

    ## Volume Type
    volume_type: VolumeType
    # Local represents directly-attached storage with node affinity
    local: Optional[LocalVolumeSource] = None
    # HostPath represents a directory on the host. Provisioned by a developer or tester.
    # This is useful for single-node development and testing only!
    # On-host storage is not supported in any way and WILL NOT WORK in a multi-node cluster.
    # More info: https://kubernetes.io/docs/concepts/storage/volumes#hostpath
    host_path: Optional[HostPathVolumeSource] = None
    # GCEPersistentDisk represents a GCE Disk resource that is attached to a
    # kubelet's host machine and then exposed to the pod. Provisioned by an admin.
    # More info: https://kubernetes.io/docs/concepts/storage/volumes#gcepersistentdisk
    gce_persistent_disk: Optional[GcePersistentDiskVolumeSource] = None

    def create(self) -> Optional[PersistentVolume]:
        """Creates the PersistentVolume resource"""

        pv_name = self.pv_name
        logger.debug(f"Init PersistentVolume resource: {pv_name}")

        pv_labels = create_component_labels_dict(
            component_name=pv_name,
            app_name=self.app_name,
            labels=self.labels,
        )
        persistent_volume = PersistentVolume(
            name=pv_name,
            api_version=ApiVersion.CORE_V1,
            kind=Kind.PERSISTENTVOLUME,
            metadata=ObjectMeta(
                name=pv_name,
                namespace=self.namespace,
                labels=pv_labels,
            ),
            spec=PersistentVolumeSpec(
                access_modes=self.access_modes,
                capacity=self.capacity,
                mount_options=self.mount_options,
                node_affinity=self.node_affinity,
                persistent_volume_reclaim_policy=self.persistent_volume_reclaim_policy,
                storage_class_name=self.storage_class_name,
                volume_mode=self.volume_mode,
            ),
        )

        if self.volume_type == VolumeType.LOCAL:
            if self.local is not None and isinstance(self.local, LocalVolumeSource):
                persistent_volume.spec.local = self.local
            else:
                print_error(
                    f"PersistentVolume {self.volume_type.value} selected but LocalVolumeSource not provided."
                )
        elif self.volume_type == VolumeType.HOST_PATH:
            if self.host_path is not None and isinstance(
                self.host_path, HostPathVolumeSource
            ):
                persistent_volume.spec.host_path = self.host_path
            else:
                print_error(
                    f"PersistentVolume {self.volume_type.value} selected but HostPathVolumeSource not provided."
                )
        elif self.volume_type == VolumeType.GCE_PERSISTENT_DISK:
            if self.gce_persistent_disk is not None and isinstance(
                self.gce_persistent_disk, GcePersistentDiskVolumeSource
            ):
                persistent_volume.spec.gce_persistent_disk = self.gce_persistent_disk
            else:
                print_error(
                    f"PersistentVolume {self.volume_type.value} selected but GcePersistentDiskVolumeSource not provided."
                )

        return persistent_volume



from typing import Optional

from phidata.utils.log import logger
from phidata.infra.aws.resource.emr.cluster import EmrCluster
from phidata.task import PythonTask, PythonTaskArgs


class DeleteEmrClusterArgs(PythonTaskArgs):
    cluster: EmrCluster


class DeleteEmrCluster(PythonTask):
    def __init__(
        self,
        cluster: EmrCluster,
        name: str = "delete_emr_cluster",
        task_id: Optional[str] = None,
        dag_id: Optional[str] = None,
        version: Optional[str] = None,
        enabled: bool = True,
    ):
        super().__init__()
        try:
            self.args: DeleteEmrClusterArgs = DeleteEmrClusterArgs(
                cluster=cluster,
                name=name,
                task_id=task_id,
                dag_id=dag_id,
                version=version,
                enabled=enabled,
                entrypoint=delete_emr_cluster,
            )
        except Exception as e:
            logger.error(f"Args for {self.__class__.__name__} are not valid")
            raise

    @property
    def cluster(self) -> EmrCluster:
        return self.args.cluster

    @cluster.setter
    def cluster(self, cluster: EmrCluster) -> None:
        if cluster is not None:
            self.args.cluster = cluster


def delete_emr_cluster(**kwargs) -> bool:

    args: DeleteEmrClusterArgs = DeleteEmrClusterArgs(**kwargs)
    # logger.debug("DeleteEmrClusterArgs: {}".format(args))

    delete_success = args.cluster.delete()
    if delete_success:
        logger.info("Delete EmrCluster: success")
    else:
        logger.error("Delete EmrCluster: failed")
    return delete_success



from typing import List, Optional
from typing_extensions import Literal

from phidata.app import PhidataApp
from phidata.app.databox import default_databox_name
from phidata.infra.base import InfraConfig
from phidata.infra.docker.args import DockerArgs
from phidata.infra.docker.manager import DockerManager
from phidata.infra.docker.resource.image import DockerImage
from phidata.infra.docker.resource.container import DockerContainer
from phidata.infra.docker.resource.volume import DockerVolume
from phidata.infra.docker.resource.group import DockerResourceGroup
from phidata.utils.log import logger


class DockerConfig(InfraConfig):
    def __init__(
        self,
        name: Optional[str] = None,
        env: Optional[Literal["dev", "stg", "prd"]] = "dev",
        version: Optional[str] = None,
        enabled: bool = True,
        network: str = "phi",
        base_url: Optional[str] = None,
        apps: Optional[List[PhidataApp]] = None,
        images: Optional[List[DockerImage]] = None,
        containers: Optional[List[DockerContainer]] = None,
        volumes: Optional[List[DockerVolume]] = None,
        resources: Optional[List[DockerResourceGroup]] = None,
        databox: str = default_databox_name,
    ):
        super().__init__()
        try:
            self.args: DockerArgs = DockerArgs(
                name=name,
                env=env,
                version=version,
                enabled=enabled,
                network=network,
                base_url=base_url,
                apps=apps,
                images=images,
                containers=containers,
                volumes=volumes,
                resources=resources,
                databox_name=databox,
            )
        except Exception as e:
            raise

    @property
    def network(self) -> Optional[str]:
        return self.args.network if self.args else None

    @property
    def base_url(self) -> Optional[str]:
        return self.args.base_url if self.args else None

    @property
    def apps(self) -> Optional[List[PhidataApp]]:
        return self.args.apps if self.args else None

    @property
    def images(self) -> Optional[List[DockerImage]]:
        return self.args.images if self.args else None

    @property
    def containers(self) -> Optional[List[DockerContainer]]:
        return self.args.containers if self.args else None

    @property
    def volumes(self) -> Optional[List[DockerVolume]]:
        return self.args.volumes if self.args else None

    @property
    def resources(self) -> Optional[List[DockerResourceGroup]]:
        return self.args.resources if self.args else None

    @property
    def databox_name(self) -> Optional[str]:
        return self.args.databox_name if self.args else None

    def apps_are_valid(self) -> bool:
        if self.apps is None:
            return False
        for _app in self.apps:
            if not isinstance(_app, PhidataApp):
                raise TypeError("Invalid App: {}".format(_app))
        return True

    def resources_are_valid(self) -> bool:
        if self.resources is None:
            return False
        for _resource in self.resources:
            if not isinstance(_resource, DockerResourceGroup):
                raise TypeError("Invalid Resource: {}".format(_resource))
        return True

    def is_valid(self) -> bool:
        return self.apps_are_valid() or self.resources_are_valid()

    def get_docker_manager(self) -> DockerManager:
        return DockerManager(docker_args=self.args)

    def get_app_by_name(self, app_name: str) -> Optional[PhidataApp]:

        if self.apps is None:
            return None

        for _app in self.apps:
            try:
                if app_name == _app.name:
                    return _app
            except Exception as e:
                logger.exception(e)
                continue
        return None



from typing import Optional

from pydantic import BaseModel


class PhidataBaseArgs(BaseModel):
    name: Optional[str] = None
    version: Optional[str] = None
    enabled: bool = True

    class Config:
        arbitrary_types_allowed = True


class PhidataBase:
    """Phidata Base Class"""

    def __init__(self) -> None:
        self.args: Optional[PhidataBaseArgs] = None

    @property
    def name(self) -> str:
        return (
            self.args.name
            if (self.args and self.args.name)
            else self.__class__.__name__
        )

    @property
    def version(self) -> Optional[str]:
        return self.args.version if self.args else None

    @property
    def enabled(self) -> bool:
        return self.args.enabled if self.args else False



from pathlib import Path
from typing import Optional, Literal, Union, List

from phidata.asset.file import File, FileType
from phidata.asset.table.sql import SqlTable
from phidata.utils.log import logger
from phidata.task import PythonTask, PythonTaskArgs


class UploadFileToSqlArgs(PythonTaskArgs):
    file: File
    sql_table: SqlTable
    # How to behave if the sql_table already exists.
    # fail: Raise a ValueError.
    # replace: Drop the table before inserting new values.
    # append: Insert new values to the existing table.
    if_exists: Optional[Literal["fail", "replace", "append"]] = None
    # Write DataFrame index as a column.
    # Uses index_label as the column name in the table.
    index: Optional[bool] = None
    # Column label for index column(s).
    # If None is given (default) and index is True, then the index names are used.
    # A sequence should be given if the DataFrame uses MultiIndex.
    index_label: Optional[Union[str, List[str]]] = None
    # Specify the number of rows in each batch to be written at a time.
    # By default, all rows will be written at once.
    chunksize: Optional[int] = None
    create_db_engine_from_conn_id: bool = True


class UploadFileToSql(PythonTask):
    def __init__(
        self,
        file: File,
        sql_table: SqlTable,
        if_exists: Optional[Literal["fail", "replace", "append"]] = None,
        index: Optional[bool] = None,
        index_label: Optional[Union[str, List[str]]] = None,
        chunksize: Optional[int] = None,
        name: str = "load_file_to_sql",
        task_id: Optional[str] = None,
        dag_id: Optional[str] = None,
        version: Optional[str] = None,
        enabled: bool = True,
    ):
        super().__init__()
        try:
            self.args: UploadFileToSqlArgs = UploadFileToSqlArgs(
                file=file,
                sql_table=sql_table,
                if_exists=if_exists,
                index=index,
                index_label=index_label,
                chunksize=chunksize,
                name=name,
                task_id=task_id,
                dag_id=dag_id,
                version=version,
                enabled=enabled,
                entrypoint=load_file_to_sql_table,
            )
        except Exception as e:
            logger.error(f"Args for {self.__class__.__name__} are not valid")
            raise

    @property
    def file(self) -> Optional[File]:
        return self.args.file

    @file.setter
    def file(self, file: File) -> None:
        if file is not None:
            self.args.file = file

    @property
    def sql_table(self) -> Optional[SqlTable]:
        return self.args.sql_table

    @sql_table.setter
    def sql_table(self, sql_table: SqlTable) -> None:
        if sql_table is not None:
            self.args.sql_table = sql_table


def load_file_to_sql_table_pandas(args: UploadFileToSqlArgs) -> bool:

    import pandas as pd

    file_to_load: Optional[File] = args.file
    if file_to_load is None:
        logger.error("File not available")
        return False
    file_path: Optional[Path] = file_to_load.file_path
    if file_path is None:
        logger.error("FilePath not available")
        return False
    file_type: Optional[FileType] = file_to_load.file_type
    if file_type is None:
        logger.error("FileType not available")
        return False
    # logger.debug("File: {}".format(file_to_load.args))

    sql_table: Optional[SqlTable] = args.sql_table
    if sql_table is None:
        logger.error("SqlTable not available")
        return False
    # logger.debug("SqlTable: {}".format(sql_table.args))

    logger.info("Reading: {}".format(file_path))
    df: Optional[pd.DataFrame] = None
    if file_type == FileType.CSV:
        df = pd.read_csv(file_path)
    elif file_type == FileType.JSON:
        df = pd.read_json(file_path)
    elif file_type == FileType.TSV:
        df = pd.read_csv(file_path, sep="\t")

    if df is not None:
        # logger.info()("DataFrame:\n{}".format(df.head()))
        logger.info("Writing to table: {}".format(sql_table.name))
        upload_success = sql_table.write_pandas_df(
            df=df,
            if_exists=args.if_exists,
            index=args.index,
            index_label=args.index_label,
            chunksize=args.chunksize,
        )
        return upload_success
    else:
        logger.info("Could not read file into DataFrame")
        return False


def load_file_to_sql_table(**kwargs) -> bool:
    args: UploadFileToSqlArgs = UploadFileToSqlArgs(**kwargs)
    return load_file_to_sql_table_pandas(args)



from typing import List, Optional

from kubernetes.client import CoreV1Api
from kubernetes.client.models.v1_namespace import V1Namespace
from kubernetes.client.models.v1_namespace_spec import V1NamespaceSpec
from kubernetes.client.models.v1_namespace_list import V1NamespaceList
from kubernetes.client.models.v1_status import V1Status

from phidata.infra.k8s.api_client import K8sApiClient
from phidata.infra.k8s.resource.base import K8sResource, K8sObject
from phidata.utils.cli_console import print_info
from phidata.utils.log import logger


class NamespaceSpec(K8sObject):
    """
    # https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#namespacespec-v1-core
    """

    resource_type: str = "NamespaceSpec"

    # Finalizers is an opaque list of values that must be empty to permanently remove object from storage.
    # More info: https://kubernetes.io/docs/tasks/administer-cluster/namespaces/
    finalizers: Optional[List[str]] = None

    def get_k8s_object(self) -> V1NamespaceSpec:

        # Return a V1NamespaceSpec object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_namespace_spec.py
        _v1_namespace_spec = V1NamespaceSpec(
            finalizers=self.finalizers,
        )
        return _v1_namespace_spec


class Namespace(K8sResource):
    """
    Kubernetes supports multiple virtual clusters backed by the same physical cluster.
    These virtual clusters are called namespaces.
    References:
        * Docs:
            https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#namespace-v1-core
            https://kubernetes.io/docs/concepts/overview/working-with-objects/namespaces/
        * Type: https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_namespace.py
    """

    resource_type: str = "Namespace"

    spec: Optional[NamespaceSpec] = None

    # List of attributes to include in the K8s manifest
    fields_for_k8s_manifest: List[str] = []

    def get_k8s_object(self) -> V1Namespace:
        """Creates a body for this Namespace"""

        # Return a V1Namespace object to create a ClusterRole
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_namespace.py
        _v1_namespace = V1Namespace(
            api_version=self.api_version,
            kind=self.kind,
            metadata=self.metadata.get_k8s_object(),
            spec=self.spec.get_k8s_object() if self.spec is not None else None,
        )
        return _v1_namespace

    @staticmethod
    def get_from_cluster(
        k8s_client: K8sApiClient, namespace: Optional[str] = None, **kwargs
    ) -> Optional[List[V1Namespace]]:
        """Reads Namespaces from K8s cluster.

        Args:
            k8s_client: K8sApiClient for the cluster
            namespace: Namespace to use.
        """
        core_v1_api: CoreV1Api = k8s_client.core_v1_api
        # logger.debug("Getting all Namespaces")
        ns_list: Optional[V1NamespaceList] = core_v1_api.list_namespace()

        namespaces: Optional[List[V1Namespace]] = None
        if ns_list:
            namespaces = [ns for ns in ns_list.items if ns.status.phase == "Active"]
        logger.debug(f"namespaces: {namespaces}")
        logger.debug(f"namespaces type: {type(namespaces)}")
        return namespaces

    def _create(self, k8s_client: K8sApiClient) -> bool:

        core_v1_api: CoreV1Api = k8s_client.core_v1_api
        k8s_object: V1Namespace = self.get_k8s_object()

        logger.debug("Creating: {}".format(self.get_resource_name()))
        v1_namespace: V1Namespace = core_v1_api.create_namespace(
            body=k8s_object,
            async_req=self.async_req,
            pretty=self.pretty,
        )
        logger.debug("Created: {}".format(v1_namespace))
        if v1_namespace.metadata.creation_timestamp is not None:
            logger.debug("Namespace Created")
            self.active_resource = v1_namespace
            self.active_resource_class = V1Namespace
            return True
        logger.error("Namespace could not be created")
        return False

    def _read(self, k8s_client: K8sApiClient) -> Optional[V1Namespace]:
        """Returns the "Active" Namespace from the cluster"""

        active_resource: Optional[V1Namespace] = None
        active_resources: Optional[List[V1Namespace]] = self.get_from_cluster(
            k8s_client=k8s_client,
        )
        # logger.debug(f"Active Resources: {active_resources}")
        if active_resources is None:
            return None

        active_resources_dict = {_ns.metadata.name: _ns for _ns in active_resources}

        ns_name = self.get_resource_name()
        if ns_name in active_resources_dict:
            active_resource = active_resources_dict[ns_name]
            self.active_resource = active_resource
            self.active_resource_class = V1Namespace
            logger.debug(f"Found active {ns_name}")
        return active_resource

    def _update(self, k8s_client: K8sApiClient) -> bool:

        core_v1_api: CoreV1Api = k8s_client.core_v1_api
        ns_name = self.get_resource_name()
        k8s_object: V1Namespace = self.get_k8s_object()

        logger.debug("Updating: {}".format(ns_name))
        v1_namespace: V1Namespace = core_v1_api.patch_namespace(
            name=ns_name,
            body=k8s_object,
            async_req=self.async_req,
            pretty=self.pretty,
        )
        # logger.debug("Updated:\n{}".format(pformat(v1_namespace.to_dict(), indent=2)))
        if v1_namespace.metadata.creation_timestamp is not None:
            logger.debug("Namespace Updated")
            self.active_resource = v1_namespace
            self.active_resource_class = V1Namespace
            return True
        logger.error("Namespace could not be updated")
        return False

    def _delete(self, k8s_client: K8sApiClient) -> bool:

        core_v1_api: CoreV1Api = k8s_client.core_v1_api
        ns_name = self.get_resource_name()

        logger.debug("Deleting: {}".format(ns_name))
        self.active_resource = None
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_status.py
        delete_status: V1Status = core_v1_api.delete_namespace(
            name=ns_name,
            async_req=self.async_req,
            pretty=self.pretty,
        )
        logger.debug("delete_status: {}".format(delete_status.status))
        if delete_status.status == "Success":
            logger.debug("Namespace Deleted")
            return True
        logger.error("Namespace could not be deleted")
        return False



from phidata.app.airflow.airflow_webserver import AirflowWebserver
from phidata.app.airflow.airflow_scheduler import AirflowScheduler
from phidata.app.airflow.airflow_worker import AirflowWorker
from phidata.app.airflow.airflow_flower import AirflowFlower



from typing import Any, List, Optional

from kubernetes.client.models.v1_config_map_env_source import V1ConfigMapEnvSource
from kubernetes.client.models.v1_config_map_key_selector import V1ConfigMapKeySelector
from kubernetes.client.models.v1_container import V1Container
from kubernetes.client.models.v1_container_port import V1ContainerPort
from kubernetes.client.models.v1_env_from_source import V1EnvFromSource
from kubernetes.client.models.v1_env_var import V1EnvVar
from kubernetes.client.models.v1_env_var_source import V1EnvVarSource
from kubernetes.client.models.v1_object_field_selector import V1ObjectFieldSelector
from kubernetes.client.models.v1_probe import V1Probe
from kubernetes.client.models.v1_resource_field_selector import V1ResourceFieldSelector
from kubernetes.client.models.v1_secret_env_source import V1SecretEnvSource
from kubernetes.client.models.v1_secret_key_selector import V1SecretKeySelector
from kubernetes.client.models.v1_volume_mount import V1VolumeMount
from pydantic import Field

from phidata.infra.k8s.enums.image_pull_policy import ImagePullPolicy
from phidata.infra.k8s.enums.protocol import Protocol
from phidata.infra.k8s.resource.base import K8sObject, K8sResource
from phidata.infra.k8s.resource.core.v1.resource_requirements import (
    ResourceRequirements,
)


class Probe(K8sObject):
    """
    Probe describes a health check to be performed against a container to determine whether it is alive or ready to receive traffic.

    # https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#probe-v1-core
    """

    resource_type: str = "Probe"

    # Minimum consecutive failures for the probe to be considered failed after having succeeded. Defaults to 3. Minimum value is 1.
    failure_threshold: Optional[int] = Field(None, alias="failureThreshold")
    # GRPC specifies an action involving a GRPC port. This is an alpha field and requires enabling GRPCContainerProbe feature gate.
    grpc: Optional[Any] = None
    # HTTPGet specifies the http request to perform.
    http_get: Optional[Any] = Field(None, alias="httpGet")
    # Number of seconds after the container has started before liveness probes are initiated. More info: https://kubernetes.io/docs/concepts/workloads/pods/pod-lifecycle#container-probes
    initial_delay_seconds: Optional[int] = Field(None, alias="initialDelaySeconds")
    # How often (in seconds) to perform the probe. Default to 10 seconds. Minimum value is 1.
    period_seconds: Optional[int] = Field(None, alias="periodSeconds")
    # Minimum consecutive successes for the probe to be considered successful after having failed. Defaults to 1. Must be 1 for liveness and startup. Minimum value is 1.
    success_threshold: Optional[int] = Field(None, alias="successThreshold")
    # TCPSocket specifies an action involving a TCP port.
    tcp_socket: Optional[Any] = Field(None, alias="tcpSocket")
    # Optional duration in seconds the pod needs to terminate gracefully upon probe failure. The grace period is the duration in seconds after the processes running in the pod are sent a termination signal and the time when the processes are forcibly halted with a kill signal. Set this value longer than the expected cleanup time for your process. If this value is nil, the pod's terminationGracePeriodSeconds will be used. Otherwise, this value overrides the value provided by the pod spec. Value must be non-negative integer. The value zero indicates stop immediately via the kill signal (no opportunity to shut down). This is a beta field and requires enabling ProbeTerminationGracePeriod feature gate. Minimum value is 1. spec.terminationGracePeriodSeconds is used if unset.
    termination_grace_period_seconds: Optional[int] = Field(
        None, alias="terminationGracePeriodSeconds"
    )
    # Number of seconds after which the probe times out. Defaults to 1 second. Minimum value is 1. More info: https://kubernetes.io/docs/concepts/workloads/pods/pod-lifecycle#container-probes
    timeout_seconds: Optional[int] = Field(None, alias="timeoutSeconds")

    def get_k8s_object(self) -> V1Probe:
        _v1_probe = V1Probe(
            failure_threshold=self.failure_threshold,
            http_get=self.http_get,
            initial_delay_seconds=self.initial_delay_seconds,
            period_seconds=self.period_seconds,
            success_threshold=self.success_threshold,
            tcp_socket=self.tcp_socket,
            termination_grace_period_seconds=self.termination_grace_period_seconds,
            timeout_seconds=self.timeout_seconds,
        )
        return _v1_probe


class ResourceFieldSelector(K8sObject):
    """
    ResourceFieldSelector represents container resources (cpu, memory) and their output format

    https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#resourcefieldselector-v1-core
    """

    resource_type: str = "ResourceFieldSelector"

    container_name: str = Field(..., alias="containerName")
    divisor: str
    resource: str

    def get_k8s_object(self) -> V1ResourceFieldSelector:

        # Return a V1ResourceFieldSelector object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_resource_field_selector.py
        _v1_resource_field_selector = V1ResourceFieldSelector(
            container_name=self.container_name,
            divisor=self.divisor,
            resource=self.resource,
        )
        return _v1_resource_field_selector


class ObjectFieldSelector(K8sObject):
    """
    ObjectFieldSelector selects an APIVersioned field of an object.

    https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#objectfieldselector-v1-core
    """

    resource_type: str = "ObjectFieldSelector"

    api_version: str = Field(..., alias="apiVersion")
    field_path: str = Field(..., alias="fieldPath")

    def get_k8s_object(self) -> V1ObjectFieldSelector:

        # Return a V1ObjectFieldSelector object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_object_field_selector.py
        _v1_object_field_selector = V1ObjectFieldSelector(
            api_version=self.api_version,
            field_path=self.field_path,
        )
        return _v1_object_field_selector


class SecretKeySelector(K8sObject):
    """
    SecretKeySelector selects a key of a Secret.

    # https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#secretkeyselector-v1-core
    """

    resource_type: str = "SecretKeySelector"

    key: str
    name: str
    optional: Optional[bool] = None

    def get_k8s_object(self) -> V1SecretKeySelector:

        # Return a V1SecretKeySelector object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_secret_key_selector.py
        _v1_secret_key_selector = V1SecretKeySelector(
            key=self.key,
            name=self.name,
            optional=self.optional,
        )
        return _v1_secret_key_selector


class ConfigMapKeySelector(K8sObject):
    """
    Selects a key from a ConfigMap.

    # https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#configmapkeyselector-v1-core
    """

    resource_type: str = "ConfigMapKeySelector"

    key: str
    name: str
    optional: Optional[bool] = None

    def get_k8s_object(self) -> V1ConfigMapKeySelector:

        # Return a V1ConfigMapKeySelector object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_config_map_key_selector.py
        _v1_config_map_key_selector = V1ConfigMapKeySelector(
            key=self.key,
            name=self.name,
            optional=self.optional,
        )
        return _v1_config_map_key_selector


class EnvVarSource(K8sObject):
    """
    EnvVarSource represents a source for the value of an EnvVar.

    # https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#envvarsource-v1-core
    """

    resource_type: str = "EnvVarSource"

    config_map_key_ref: Optional[ConfigMapKeySelector] = Field(
        None, alias="configMapKeyRef"
    )
    field_ref: Optional[ObjectFieldSelector] = Field(None, alias="fieldRef")
    resource_field_ref: Optional[ResourceFieldSelector] = Field(
        None, alias="resourceFieldRef"
    )
    secret_key_ref: Optional[SecretKeySelector] = Field(None, alias="secretKeyRef")

    def get_k8s_object(self) -> V1EnvVarSource:

        # Return a V1EnvVarSource object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_env_var_source.py
        _v1_env_var_source = V1EnvVarSource(
            config_map_key_ref=self.config_map_key_ref.get_k8s_object()
            if self.config_map_key_ref
            else None,
            field_ref=self.field_ref.get_k8s_object() if self.field_ref else None,
            resource_field_ref=self.resource_field_ref.get_k8s_object()
            if self.resource_field_ref
            else None,
            secret_key_ref=self.secret_key_ref.get_k8s_object()
            if self.secret_key_ref
            else None,
        )
        return _v1_env_var_source


class EnvVar(K8sObject):
    """
    EnvVar represents an environment variable present in a Container.

    # https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#envvar-v1-core
    """

    resource_type: str = "EnvVar"

    name: str
    value: Optional[str] = None
    value_from: Optional[EnvVarSource] = Field(None, alias="valueFrom")

    def get_k8s_object(self) -> V1EnvVar:

        # Return a V1EnvVar object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_env_var.py
        _v1_env_var = V1EnvVar(
            name=self.name,
            value=self.value,
            value_from=self.value_from.get_k8s_object() if self.value_from else None,
        )
        return _v1_env_var


class ConfigMapEnvSource(K8sObject):
    """
    ConfigMapEnvSource selects a ConfigMap to populate the environment variables with.
    The contents of the target ConfigMap's Data field will represent the key-value pairs as environment variables.

    # https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#configmapenvsource-v1-core
    """

    resource_type: str = "ConfigMapEnvSource"

    name: str
    optional: Optional[bool] = None

    def get_k8s_object(self) -> V1ConfigMapEnvSource:
        _v1_config_map_env_source = V1ConfigMapEnvSource(
            name=self.name,
            optional=self.optional,
        )
        return _v1_config_map_env_source


class SecretEnvSource(K8sObject):
    """
    SecretEnvSource selects a Secret to populate the environment variables with.
    The contents of the target Secret's Data field will represent the key-value pairs as environment variables.

    # https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#secretenvsource-v1-core
    """

    resource_type: str = "SecretEnvSource"

    name: str
    optional: Optional[bool] = None

    def get_k8s_object(self) -> V1SecretEnvSource:
        _v1_secret_env_source = V1SecretEnvSource(
            name=self.name,
            optional=self.optional,
        )
        return _v1_secret_env_source


class EnvFromSource(K8sObject):
    """
    EnvFromSource represents the source of a set of ConfigMaps

    # https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#envfromsource-v1-core
    """

    resource_type: str = "EnvFromSource"

    config_map_ref: Optional[ConfigMapEnvSource] = Field(None, alias="configMapRef")
    prefix: Optional[str] = None
    secret_ref: Optional[SecretEnvSource] = Field(None, alias="secretRef")

    def get_k8s_object(self) -> V1EnvFromSource:

        # Return a V1EnvFromSource object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_env_from_source.py
        _v1_env_from_source = V1EnvFromSource(
            config_map_ref=self.config_map_ref.get_k8s_object()
            if self.config_map_ref
            else None,
            prefix=self.prefix,
            secret_ref=self.secret_ref.get_k8s_object() if self.secret_ref else None,
        )
        return _v1_env_from_source


class ContainerPort(K8sObject):
    """
    ContainerPort represents a network port in a single container.

    # https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#containerport-v1-core
    """

    resource_type: str = "ContainerPort"

    # If specified, this must be an IANA_SVC_NAME and unique within the pod.
    # Each named port in a pod must have a unique name.
    # Name for the port that can be referred to by services.
    name: str
    # Number of port to expose on the pod's IP address. This must be a valid port number, 0 < x < 65536.
    container_port: int = Field(..., alias="containerPort")
    host_ip: Optional[str] = Field(None, alias="hostIP")
    # Number of port to expose on the host.
    # If specified, this must be a valid port number, 0 < x < 65536.
    # If HostNetwork is specified, this must match ContainerPort.
    # Most containers do not need this.
    host_port: Optional[int] = Field(None, alias="hostPort")
    protocol: Optional[Protocol] = None

    def get_k8s_object(self) -> V1ContainerPort:
        _v1_container_port = V1ContainerPort(
            container_port=self.container_port,
            name=self.name,
            protocol=self.protocol,
            host_ip=self.host_ip,
            host_port=self.host_port,
        )
        return _v1_container_port


class VolumeMount(K8sObject):
    """
    VolumeMount describes a mounting of a Volume within a container.

    # https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#volumemount-v1-core
    """

    resource_type: str = "VolumeMount"

    # Path within the container at which the volume should be mounted. Must not contain ':'
    mount_path: str = Field(..., alias="mountPath")
    # mountPropagation determines how mounts are propagated from the host to container and the other way around. When not set, MountPropagationNone is used. This field is beta in 1.10.
    mount_propagation: Optional[str] = Field(None, alias="mountPropagation")
    # This must match the Name of a Volume.
    name: str
    # Mounted read-only if true, read-write otherwise (false or unspecified). Defaults to false.
    read_only: Optional[bool] = Field(None, alias="readOnly")
    # Path within the volume from which the container's volume should be mounted. Defaults to "" (volume's root).
    sub_path: Optional[str] = Field(None, alias="subPath")
    # Expanded path within the volume from which the container's volume should be mounted.
    # Behaves similarly to SubPath but environment variable references $(VAR_NAME) are expanded using the container's environment.
    # Defaults to "" (volume's root). SubPathExpr and SubPath are mutually exclusive.
    sub_path_expr: Optional[str] = Field(None, alias="subPathExpr")

    def get_k8s_object(self) -> V1VolumeMount:
        _v1_volume_mount = V1VolumeMount(
            mount_path=self.mount_path,
            mount_propagation=self.mount_propagation,
            name=self.name,
            read_only=self.read_only,
            sub_path=self.sub_path,
            sub_path_expr=self.sub_path_expr,
        )
        return _v1_volume_mount


class Container(K8sObject):
    """
    # https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#container-v1-core
    """

    resource_type: str = "Container"

    # Arguments to the entrypoint. The docker image's CMD is used if this is not provided.
    args: Optional[List[str]] = None
    # Entrypoint array. Not executed within a shell. The docker image's ENTRYPOINT is used if this is not provided.
    command: Optional[List[str]] = None
    env: Optional[List[EnvVar]] = None
    env_from: Optional[List[EnvFromSource]] = Field(None, alias="envFrom")
    # Docker image name.
    image: str
    # Image pull policy. One of Always, Never, IfNotPresent.
    # Defaults to Always if :latest tag is specified, or IfNotPresent otherwise.
    image_pull_policy: Optional[ImagePullPolicy] = Field(None, alias="imagePullPolicy")
    # Name of the container specified as a DNS_LABEL.
    # Each container in a pod must have a unique name (DNS_LABEL).
    name: str
    # List of ports to expose from the container.
    # Exposing a port here gives the system additional information about the network connections a container uses, but is primarily informational.
    # Not specifying a port here DOES NOT prevent that port from being exposed.
    ports: Optional[List[ContainerPort]] = None
    # TODO: add Probe object
    # Periodic probe of container service readiness.
    # Container will be removed from service endpoints if the probe fails. Cannot be updated.
    readiness_probe: Optional[Probe] = Field(None, alias="readinessProbe")
    # Compute Resources required by this container. Cannot be updated.
    resources: Optional[ResourceRequirements] = None
    volume_mounts: Optional[List[VolumeMount]] = Field(None, alias="volumeMounts")
    working_dir: Optional[str] = Field(None, alias="workingDir")

    def get_k8s_object(self) -> V1Container:

        # Return a V1Container object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_container.py
        _ports: Optional[List[V1ContainerPort]] = None
        if self.ports:
            _ports = [_cp.get_k8s_object() for _cp in self.ports]

        _env: Optional[List[V1EnvVar]] = None
        if self.env:
            _env = [_e.get_k8s_object() for _e in self.env]

        _env_from: Optional[List[V1EnvFromSource]] = None
        if self.env_from:
            _env_from = [_ef.get_k8s_object() for _ef in self.env_from]

        _volume_mounts: Optional[List[V1VolumeMount]] = None
        if self.volume_mounts:
            _volume_mounts = [_vm.get_k8s_object() for _vm in self.volume_mounts]

        _v1_container = V1Container(
            args=self.args,
            command=self.command,
            env=_env,
            env_from=_env_from,
            image=self.image,
            image_pull_policy=self.image_pull_policy,
            name=self.name,
            ports=_ports,
            readiness_probe=self.readiness_probe.get_k8s_object()
            if self.readiness_probe
            else None,
            resources=self.resources.get_k8s_object() if self.resources else None,
            volume_mounts=_volume_mounts,
        )
        return _v1_container



from collections import defaultdict, OrderedDict
from typing import Any, Dict, List, Optional, Type, Set, cast

from phidata.infra.k8s.args import K8sArgs
from phidata.infra.k8s.constants import (
    DEFAULT_K8S_NAMESPACE,
    DEFAULT_K8S_SERVICE_ACCOUNT,
    NAMESPACE_RESOURCE_GROUP_KEY,
    RBAC_RESOURCE_GROUP_KEY,
)
from phidata.infra.k8s.api_client import K8sApiClient
from phidata.infra.k8s.exceptions import K8sArgsException
from phidata.infra.k8s.resource.base import K8sResource
from phidata.infra.k8s.create.group import CreateK8sResourceGroup
from phidata.infra.k8s.resource.group import (
    K8sResourceGroup,
    K8sBuildContext,
)
from phidata.infra.k8s.resource.kubeconfig import Kubeconfig
from phidata.infra.k8s.resource.utils import (
    filter_and_flatten_k8s_resource_groups,
    dedup_resource_types,
)
from phidata.utils.cli_console import (
    print_info,
    print_info,
    print_heading,
    print_subheading,
    print_error,
    confirm_yes_no,
)
from phidata.utils.common import is_empty
from phidata.utils.log import logger


class K8sWorker:
    """This class interacts with the Kubernetes API."""

    def __init__(self, k8s_args: K8sArgs) -> None:
        # logger.debug(f"Creating K8sWorker")
        if k8s_args is None or not isinstance(k8s_args, K8sArgs):
            raise K8sArgsException("K8sArgs invalid: {}".format(k8s_args))

        self.k8s_args: K8sArgs = k8s_args
        self.k8s_api_client: K8sApiClient = K8sApiClient(
            context=self.k8s_args.context,
            kubeconfig_path=self.k8s_args.kubeconfig_path,
        )
        self.k8s_resources: Optional[Dict[str, K8sResourceGroup]] = None
        logger.debug(f"**-+-** K8sWorker created")

    def is_client_initialized(self) -> bool:
        return self.k8s_api_client.is_initialized()

    def are_resources_initialized(self) -> bool:
        if self.k8s_resources is not None and len(self.k8s_resources) > 0:
            return True
        return False

    ######################################################
    ## Init Resources
    ######################################################

    def init_resources(self) -> None:
        """
        This function populates the self.k8s_resources dictionary.

        Step 1: Convert each PhidataApp to K8sResourceGroups.
        Step 2: Convert each CreateK8sResourceGroup to K8sResourceGroups.
        Step 3: Add all K8sResourceGroups to the self.k8s_resources dict
        """
        logger.debug("-*- Initializing K8sResourceGroups")

        # track the total number of K8sResourceGroups to build for validation
        apps_to_build = self.k8s_args.apps
        resources_to_build = self.k8s_args.resources
        create_resources_to_build = self.k8s_args.create_resources

        num_apps = len(apps_to_build) if apps_to_build is not None else 0
        num_resources = len(resources_to_build) if resources_to_build is not None else 0
        num_create_resources = (
            len(create_resources_to_build)
            if create_resources_to_build is not None
            else 0
        )
        num_rgs_to_build = num_apps + num_resources + num_create_resources
        num_rgs_built = 0

        # Step 1: Convert each PhidataApp to K8sResourceGroups.
        if apps_to_build is not None and isinstance(apps_to_build, list):
            for app in apps_to_build:
                if app.args is None:
                    print_error("Args for App {} are None".format(app))
                    continue
                logger.debug("-*- App: {}".format(app.name))

                ######################################################################
                # NOTE: VERY IMPORTANT TO GET RIGHT
                # Pass down the paths and env
                # from  WorkspaceConfig -> DockerConfig -> PhidataApp
                ######################################################################

                app.workspace_root_path = self.k8s_args.workspace_root_path
                app.workspace_config_file_path = (
                    self.k8s_args.workspace_config_file_path
                )
                app.scripts_dir = self.k8s_args.scripts_dir
                app.storage_dir = self.k8s_args.storage_dir
                app.meta_dir = self.k8s_args.meta_dir
                app.products_dir = self.k8s_args.products_dir
                app.notebooks_dir = self.k8s_args.notebooks_dir
                app.workspace_config_dir = self.k8s_args.workspace_config_dir
                app.k8s_env = self.k8s_args.k8s_env

                app_rgs: Optional[
                    Dict[str, K8sResourceGroup]
                ] = app.get_k8s_resource_groups(
                    k8s_build_context=K8sBuildContext(
                        namespace=self.k8s_args.namespace,
                        context=self.k8s_args.context,
                        service_account_name=self.k8s_args.service_account_name,
                        labels=self.k8s_args.labels,
                    )
                )
                if app_rgs is not None:
                    if self.k8s_resources is None:
                        self.k8s_resources = OrderedDict()
                    self.k8s_resources.update(app_rgs)
                    num_rgs_built += 1

        # Step 2: Convert each CreateK8sResourceGroup to K8sResourceGroups.
        if create_resources_to_build is not None and isinstance(
            create_resources_to_build, list
        ):
            for create_resource in create_resources_to_build:
                if not create_resource.enabled:
                    print_info(f"{create_resource.name} disabled")
                    continue
                logger.debug("-*- Resource: {}".format(create_resource.name))

                if isinstance(create_resource, CreateK8sResourceGroup):
                    # Create the K8sResourceGroup
                    k8s_resource_group: Optional[
                        K8sResourceGroup
                    ] = create_resource.create()
                    if k8s_resource_group is not None:
                        if self.k8s_resources is None:
                            self.k8s_resources = OrderedDict()
                        self.k8s_resources[k8s_resource_group.name] = k8s_resource_group
                    num_rgs_built += 1

        # Step 3: Add all K8sResourceGroups to the self.k8s_resources dict
        if resources_to_build is not None and isinstance(resources_to_build, list):
            for resource in resources_to_build:
                if not resource.enabled:
                    print_info(f"{resource.name} disabled")
                    continue
                logger.debug("-*- Resource: {}".format(resource.name))

                if isinstance(resource, K8sResourceGroup):
                    if self.k8s_resources is None:
                        self.k8s_resources = OrderedDict()
                    self.k8s_resources[resource.name] = resource
                    num_rgs_built += 1

        logger.debug(f"# K8sResourceGroups built: {num_rgs_built}/{num_rgs_to_build}")

    ######################################################
    ## Create Resources
    ######################################################

    def create_resources(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
    ) -> bool:
        logger.debug("-*- Creating K8sResources")
        if self.k8s_resources is None:
            self.init_resources()
            if self.k8s_resources is None:
                print_info("No Resources available")
                return False

        k8s_resources_to_create: Optional[
            List[K8sResource]
        ] = filter_and_flatten_k8s_resource_groups(
            k8s_resource_groups=self.k8s_resources,
            name_filter=name_filter,
            type_filter=type_filter,
            sort_order="create",
        )
        if k8s_resources_to_create is None or is_empty(k8s_resources_to_create):
            print_info("No K8sResources to create")
            return True

        # track the total number of K8sResources to create for validation
        num_resources_to_create: int = len(k8s_resources_to_create)
        num_resources_created: int = 0

        # Print the resources for validation
        print_subheading(f"Creating {num_resources_to_create} K8sResources:")
        for rsrc in k8s_resources_to_create:
            print_info(f"  -+-> {rsrc.get_resource_type()}: {rsrc.get_resource_name()}")
        print_info(f"\n  namespace: {self.k8s_args.namespace}")
        print_info(f"  context: {self.k8s_args.context}")
        confirm = confirm_yes_no("\nConfirm deploy")
        if not confirm:
            print_info("Skipping deploy")
            return False

        for resource in k8s_resources_to_create:
            if resource:
                print_info(
                    f"\n-==+==- {resource.get_resource_type()}: {resource.get_resource_name()}"
                )
                # logger.debug(resource)
                try:
                    _resource_created = resource.create(k8s_client=self.k8s_api_client)
                    if _resource_created:
                        num_resources_created += 1
                except Exception as e:
                    print_error(
                        f"-==+==--> Resource {resource.get_resource_type()}: {resource.get_resource_name()} could not be created."
                    )
                    print_error("Error: {}".format(e))
                    print_error(
                        "Skipping resource creation, please fix and try again..."
                    )

        print_info(
            f"\n# Resources created: {num_resources_created}/{num_resources_to_create}"
        )
        if num_resources_to_create == num_resources_created:
            return True

        print_error(
            f"Resources created: {num_resources_created} do not match Resources required: {num_resources_to_create}"
        )
        return False

    def create_resources_dry_run(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
    ) -> None:

        env = self.k8s_args.env
        print_subheading(
            "K8sResources{}".format(f" for env: {env}" if env is not None else "")
        )
        if self.k8s_resources is None:
            self.init_resources()
            if self.k8s_resources is None:
                print_info("No Resources available")
                return

        k8s_resources_to_create: Optional[
            List[K8sResource]
        ] = filter_and_flatten_k8s_resource_groups(
            k8s_resource_groups=self.k8s_resources,
            name_filter=name_filter,
            type_filter=type_filter,
            sort_order="create",
        )
        if k8s_resources_to_create is None or is_empty(k8s_resources_to_create):
            print_info("No K8sResources to create")
            return

        num_resources_to_create: int = len(k8s_resources_to_create)
        print_info(
            f"\n-*- Deploying this K8sConfig will create {num_resources_to_create} resources:"
        )
        print_info(f"\tnamespace: {self.k8s_args.namespace}")
        print_info(f"\tcontext: {self.k8s_args.context}")
        for resource in k8s_resources_to_create:
            print_info(
                f"-==+==- {resource.__class__.__name__}: {resource.get_resource_name()}"
            )

    ######################################################
    ## Read Resources
    ######################################################

    def get_resources(self) -> Optional[Dict[str, K8sResourceGroup]]:

        logger.debug("-*- Getting K8sResources")
        if self.k8s_resources is None:
            self.init_resources()
            if self.k8s_resources is None:
                print_info("No Resources available")
                return {}

        return self.k8s_resources

    def read_resources(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
    ) -> Optional[List[K8sResource]]:

        logger.debug("-*- Getting K8sResources")
        if self.k8s_resources is None:
            self.init_resources()
            if self.k8s_resources is None:
                print_info("No Resources available")
                return None

        k8s_resources: Optional[
            List[K8sResource]
        ] = filter_and_flatten_k8s_resource_groups(
            k8s_resource_groups=self.k8s_resources,
            name_filter=name_filter,
            type_filter=type_filter,
        )
        return k8s_resources

    ######################################################
    ## Delete Resources
    ######################################################

    def delete_resources(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
    ) -> bool:

        logger.debug("-*- Deleting K8sResources")
        if self.k8s_resources is None:
            self.init_resources()
            if self.k8s_resources is None:
                print_info("No Resources available")
                return False

        k8s_resources_to_delete: Optional[
            List[K8sResource]
        ] = filter_and_flatten_k8s_resource_groups(
            k8s_resource_groups=self.k8s_resources,
            name_filter=name_filter,
            type_filter=type_filter,
            sort_order="delete",
        )
        if k8s_resources_to_delete is None or is_empty(k8s_resources_to_delete):
            print_info("No K8sResources to delete")
            return True

        # track the total number of K8sResources to delete for validation
        num_resources_to_delete: int = len(k8s_resources_to_delete)
        num_resources_deleted: int = 0

        # Print the resources for validation
        # Print the resources for validation
        print_subheading(f"Deleting {num_resources_to_delete} K8sResources:")
        for rsrc in k8s_resources_to_delete:
            print_info(f"  -+-> {rsrc.get_resource_type()}: {rsrc.get_resource_name()}")
        confirm = confirm_yes_no("\nConfirm delete")
        if not confirm:
            print_info("Skipping delete")
            return False

        for resource in k8s_resources_to_delete:
            if resource:
                print_info(
                    f"\n-==+==- {resource.get_resource_type()}: {resource.get_resource_name()}"
                )
                # logger.debug(resource)
                try:
                    _resource_deleted = resource.delete(k8s_client=self.k8s_api_client)
                    if _resource_deleted:
                        num_resources_deleted += 1
                except Exception as e:
                    print_error(
                        f"-==+==--> Resource {resource.get_resource_type()}: {resource.get_resource_name()} could not be deleted."
                    )
                    print_error(e)
                    print_error("Skipping resource deletion, please delete manually...")

        print_info(
            f"\n# Resources deleted: {num_resources_deleted}/{num_resources_to_delete}"
        )
        if num_resources_to_delete == num_resources_deleted:
            return True

        print_error(
            f"Resources deleted: {num_resources_deleted} do not match Resources required: {num_resources_to_delete}"
        )
        return False

    def delete_resources_dry_run(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
    ) -> None:

        env = self.k8s_args.env
        print_subheading(
            "K8sResources{}".format(f" for env: {env}" if env is not None else "")
        )
        if self.k8s_resources is None:
            self.init_resources()
            if self.k8s_resources is None:
                print_info("No Resources available")
                return

        k8s_resources_to_delete: Optional[
            List[K8sResource]
        ] = filter_and_flatten_k8s_resource_groups(
            k8s_resource_groups=self.k8s_resources,
            name_filter=name_filter,
            type_filter=type_filter,
            sort_order="delete",
        )
        if k8s_resources_to_delete is None or is_empty(k8s_resources_to_delete):
            print_info("No K8sResources to delete")
            return

        num_resources_to_delete: int = len(k8s_resources_to_delete)
        print_info(
            f"\n-*- Shutting down this K8sConfig will delete {num_resources_to_delete} resources:"
        )
        for resource in k8s_resources_to_delete:
            print_info(
                f"-==+==- {resource.__class__.__name__}: {resource.get_resource_name()}"
            )

    ######################################################
    ## Patch Resources
    ######################################################

    def patch_resources(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
    ) -> bool:

        logger.debug("-*- Patching K8sResources")
        if self.k8s_resources is None:
            self.init_resources()
            if self.k8s_resources is None:
                print_info("No Resources available")
                return False

        k8s_resources_to_patch: Optional[
            List[K8sResource]
        ] = filter_and_flatten_k8s_resource_groups(
            k8s_resource_groups=self.k8s_resources,
            name_filter=name_filter,
            type_filter=type_filter,
            sort_order="create",
        )
        if k8s_resources_to_patch is None or is_empty(k8s_resources_to_patch):
            print_info("No K8sResources to patch")
            return True

        # track the total number of K8sResources to patch for validation
        num_resources_to_patch: int = len(k8s_resources_to_patch)
        num_resources_patched: int = 0

        # Print the resources for validation
        print_subheading(f"Patching {num_resources_to_patch} K8sResources:")
        for rsrc in k8s_resources_to_patch:
            print_info(f"  -+-> {rsrc.get_resource_type()}: {rsrc.get_resource_name()}")
        confirm = confirm_yes_no("\nConfirm patch")
        if not confirm:
            print_info("Skipping patch")
            return False

        for resource in k8s_resources_to_patch:
            if resource:
                print_info(
                    f"\n-==+==- {resource.get_resource_type()}: {resource.get_resource_name()}"
                )
                # logger.debug(resource)
                try:
                    _resource_patched = resource.update(k8s_client=self.k8s_api_client)
                    if _resource_patched:
                        num_resources_patched += 1
                except Exception as e:
                    print_error(
                        f"-==+==--> Resource {resource.get_resource_type()}: {resource.get_resource_name()} could not be patched."
                    )
                    print_error(e)
                    print_error("Skipping resource, please patch manually...")

        print_info(
            f"\n# Resources patched: {num_resources_patched}/{num_resources_to_patch}"
        )
        if num_resources_to_patch == num_resources_patched:
            return True

        print_error(
            f"Resources patched: {num_resources_patched} do not match Resources required: {num_resources_to_patch}"
        )
        return False

    def patch_resources_dry_run(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
    ) -> None:

        env = self.k8s_args.env
        print_subheading(
            "K8sResources{}".format(f" for env: {env}" if env is not None else "")
        )
        if self.k8s_resources is None:
            self.init_resources()
            if self.k8s_resources is None:
                print_info("No Resources available")
                return

        k8s_resources_to_patch: Optional[
            List[K8sResource]
        ] = filter_and_flatten_k8s_resource_groups(
            k8s_resource_groups=self.k8s_resources,
            name_filter=name_filter,
            type_filter=type_filter,
            sort_order="create",
        )
        if k8s_resources_to_patch is None or is_empty(k8s_resources_to_patch):
            print_info("No K8sResources to patch")
            return

        num_resources_to_patch: int = len(k8s_resources_to_patch)
        print_info(
            f"\n-*- Updating this K8sConfig will patch {num_resources_to_patch} resources:"
        )
        for resource in k8s_resources_to_patch:
            print_info(
                f"-==+==- {resource.__class__.__name__}: {resource.get_resource_name()}"
            )

    ######################################################
    ## End
    ######################################################

    ######################################################
    ## List Resources
    ######################################################

    # def get_active_resource_classes(
    #     self,
    #     type_filters: Optional[List[str]] = None,
    #     name_filters: Optional[List[str]] = None,
    # ) -> Optional[Set[Type[K8sResource]]]:
    #     """This function scans all K8sResourceGroups under this client and
    #     returns a set of different types of K8sResource(s) as a set of classes.
    #
    #     When do we need this:
    #     Let's say your app has 1 Deployment, 1 Pod and 1 Service. You deploy it and now want
    #     to see the different types of objects running on your cluster on your cluster.
    #     To do that you need to query the API for each and every resource type to see if there are any running objects of that type.
    #     We use this function to filter & dedup the resource types this K8s client is managing.
    #     This way we don't end up scanning everything.
    #
    #     Also super useful when you want to just see Deployments running, faster responses.
    #     """
    #     if self.k8s_resources is None:
    #         logger.debug("No K8sResourceGroups available")
    #         return None
    #
    #     _filtered_k8s_resources: Optional[
    #         List[K8sResource]
    #     ] = filter_and_flatten_k8s_resource_groups(
    #         k8s_resource_groups=self.k8s_resources,
    #         type_filters=type_filters,
    #         name_filters=name_filters,
    #         sort_order="create",
    #     )
    #     return dedup_resource_types(_filtered_k8s_resources)
    #
    # def get_active_resources(
    #     self,
    #     type_filters: Optional[List[str]] = None,
    #     name_filters: Optional[List[str]] = None,
    # ) -> Optional[Dict[str, List]]:
    #     """Reads the K8s Cluster and returns all active k8s_resources which satisfy the filters."""
    #
    #     from kubernetes.client.rest import ApiException
    #
    #     active_resource_classes: Optional[
    #         Set[Type[K8sResource]]
    #     ] = self.get_active_resource_classes(type_filters, name_filters)
    #
    #     if active_resource_classes is None:
    #         return None
    #
    #     _active_ns = self.get_k8s_namespace_to_use()
    #     active_k8s_objects: Dict[str, List] = defaultdict(list)
    #     for resource_class in active_resource_classes:
    #         resource_type: str = resource_class.__name__
    #         logger.debug(f"Resource Type: {resource_type}")
    #         try:
    #             _active_objects: Optional[List[Any]] = resource_class.get_from_cluster(
    #                 k8s_client=cast(K8sApiClient, self.k8s_api_client), namespace=_active_ns
    #             )
    #             if _active_objects is not None and isinstance(_active_objects, list):
    #                 active_k8s_objects[resource_type] = _active_objects
    #         except ApiException as e:
    #             logger.debug(
    #                 f"ApiException while getting {resource_type}, reason: {e.reason}"
    #             )
    #
    #     return active_k8s_objects
    #
    # def print(
    #     self,
    #     type_filters: Optional[List[str]] = None,
    #     name_filters: Optional[List[str]] = None,
    #     as_yaml: Optional[bool] = True,
    # ) -> None:
    #
    #     active_k8s_objects: Optional[Dict[str, List]] = self.get_active_resources(
    #         type_filters, name_filters
    #     )
    #     if active_k8s_objects:
    #         for _r_type, active_resources in active_k8s_objects.items():
    #             print("\n{}:".format(_r_type))
    #             for _r in active_resources:
    #                 print("\t- {}".format(_r))

    ######################################################
    ## Debug
    ######################################################

    # def debug(
    #     self,
    #     type_filters: Optional[List[str]] = None,
    #     name_filters: Optional[List[str]] = None,
    # ) -> None:
    #
    #     if self._k8s_client is None:
    #         logger.debug("No K8sApiClient available")
    #         return
    #     if self.k8s_resources is None:
    #         logger.debug("No K8sResourceGroups available")
    #         return
    #     _filtered_k8s_resources: Optional[
    #         List[K8sResource]
    #     ] = filter_and_flatten_k8s_resource_groups(
    #         k8s_resource_groups=self.k8s_resources,
    #         type_filters=type_filters,
    #         name_filters=name_filters,
    #         sort_order="create",
    #     )
    #     if _filtered_k8s_resources:
    #         for resource in _filtered_k8s_resources:
    #             logger.debug(f"Resource: {resource.metadata.name}")
    #             if resource and self.k8s_api_client:
    #                 resource.debug()

    ######################################################
    ## K8sApiClient
    ######################################################

    # def get_k8s_namespace_to_use(self) -> str:
    #     if self.k8s_resources is None:
    #         return DEFAULT_K8S_NAMESPACE
    #     ns_rg: K8sResourceGroup = self.k8s_resources[
    #         NAMESPACE_RESOURCE_GROUP_KEY
    #     ]
    #     if ns_rg and ns_rg.enabled and ns_rg.ns is not None:
    #         return ns_rg.ns.get_resource_name()
    #     return DEFAULT_K8S_NAMESPACE
    #
    # def get_k8s_service_account_to_use(self) -> str:
    #     if self.k8s_resources is None:
    #         return DEFAULT_K8S_SERVICE_ACCOUNT
    #     rbac_rg: K8sResourceGroup = self.k8s_resources[RBAC_RESOURCE_GROUP_KEY]
    #     if rbac_rg and rbac_rg.enabled and rbac_rg.sa is not None:
    #         return rbac_rg.sa.get_resource_name()
    #     return DEFAULT_K8S_SERVICE_ACCOUNT

    ######################################################
    ## Describe Resources
    ######################################################

    # def print_k8s_resources(
    #     self,
    #     type_filters: Optional[List[str]] = None,
    #     name_filters: Optional[List[str]] = None,
    #     op_format: str = "yaml",  # yaml or json
    # ) -> None:
    #
    #     if op_format not in ("yaml", "json"):
    #         print(f"Output format {op_format} not supported")
    #         return
    #
    #     if self.k8s_resources is None:
    #         logger.debug("No K8sResourceGroups available")
    #         return
    #
    #     self.print_k8s_resource_groups(
    #         k8s_resource_groups=self.k8s_resources,
    #         type_filters=type_filters,
    #         name_filters=name_filters,
    #         as_yaml=(op_format == "yaml"),
    #     )
    #
    # @staticmethod
    # def print_k8s_resource_groups(
    #     k8s_resource_groups: Dict[str, K8sResourceGroup],
    #     type_filters: Optional[List[str]] = None,
    #     name_filters: Optional[List[str]] = None,
    #     as_yaml: Optional[bool] = True,
    # ) -> None:
    #
    #     print(
    #         "Resources: {}/{}".format(
    #             type_filters if type_filters else "*",
    #             name_filters if name_filters else "*",
    #         )
    #     )
    #     _filtered_k8s_resources: Optional[
    #         List[K8sResource]
    #     ] = filter_and_flatten_k8s_resource_groups(
    #         k8s_resource_groups=k8s_resource_groups,
    #         type_filters=type_filters,
    #         name_filters=name_filters,
    #         sort_order="create",
    #     )
    #     _resources: List[str] = []
    #     if _filtered_k8s_resources:
    #         for resource in _filtered_k8s_resources:
    #             if resource:
    #                 if as_yaml:
    #                     _yml = resource.get_k8s_manifest_yaml()
    #                     if _yml:
    #                         _resources.append(_yml)
    #                 else:
    #                     _json = resource.get_k8s_manifest_json(indent=2)
    #                     if _json:
    #                         _resources.append(_json)
    #
    #     if _resources:
    #         if as_yaml:
    #             print("---")
    #             print("---\n".join(_resources))
    #         else:
    #             print("\n".join(_resources))

    ######################################################
    ## Helpers
    ######################################################

    # def _debug_log(self, msg: Optional[str] = None) -> None:
    #     if msg:
    #         logger.debug(msg)
    #     kubeconfig_avl = True if self._kubeconfig else False
    #     k8s_client_avl = True if self.k8s_api_client else False
    #     logger.debug(f"Kubeconfig Available  : {kubeconfig_avl}")
    #     logger.debug(f"K8s_api Available  : {k8s_client_avl}")
    #
    # def verify_client(self) -> None:
    #     """Helper method to verify that we are good to perform K8s operaitons.
    #
    #     Raises:
    #         K8sClientException if something is wrong
    #     """
    #     if self._kubeconfig and self._k8s_client:
    #         pass
    #     else:
    #         self._debug_log()
    #         raise exceptions.K8sWorkerException("K8sWorker unavailable")



from typing import List, Optional

from pydantic import BaseModel

from phidata.infra.docker.resource.network import DockerNetwork
from phidata.infra.docker.resource.image import DockerImage
from phidata.infra.docker.resource.container import DockerContainer
from phidata.infra.docker.resource.volume import DockerVolume


class DockerResourceGroup(BaseModel):
    """The DockerResourceGroup class contains the instructions to manage docker resources"""

    name: str
    enabled: bool = True

    # The weight variable controls how this is resource group is deployed to a cluster
    # relative to other resource groups.

    # Within each resource group, the resources are deployed in a predefined order
    # eg: network first, then images, volumes, containers and so on..
    # But we can also add an order to how different resource groups are deployed relative to each other.
    # (Eg if we want to deploy a resource group with just storage_class (s) before all other resources)

    # Weights 1-10 are reserved
    # Weight 100 is default.
    # Choose weight 11-99 to deploy a resource group before all the default resources.
    # Choose weight 101+ to deploy a resource group after all the default resources
    weight: int = 100
    network: Optional[DockerNetwork] = None
    images: Optional[List[DockerImage]] = None
    containers: Optional[List[DockerContainer]] = None
    # secrets: Optional[List[DockerSecret]] = None
    volumes: Optional[List[DockerVolume]] = None


class DockerBuildContext(BaseModel):
    """This class is used to store default information when creating a DockerResourceGroup"""

    network: str



from pathlib import Path
from typing import List, Optional, Dict

from pydantic import validator

from phidata.app import PhidataApp
from phidata.app.databox import default_databox_name
from phidata.infra.aws.resource.eks.cluster import EksCluster
from phidata.infra.base import InfraArgs
from phidata.infra.k8s.create.group import CreateK8sResourceGroup
from phidata.infra.k8s.resource.group import K8sResourceGroup


class K8sArgs(InfraArgs):
    # K8s namespace to use
    namespace: str = "default"
    # K8s context to use
    context: Optional[str] = None
    service_account_name: Optional[str] = None
    labels: Optional[Dict[str, str]] = None
    apps: Optional[List[PhidataApp]] = None
    resources: Optional[List[K8sResourceGroup]] = None
    create_resources: Optional[List[CreateK8sResourceGroup]] = None
    resources_dir: str = "k8s"
    databox_name: str = default_databox_name
    kubeconfig_path: Optional[Path] = None

    # Get K8s context and kubeconfig from
    eks_cluster: Optional[EksCluster] = None

    @validator("apps")
    def apps_are_valid(cls, apps):
        if apps is not None:
            for _app in apps:
                if not isinstance(_app, PhidataApp):
                    raise TypeError("App not of type PhidataApp: {}".format(_app))
            return apps



from typing import Optional, List, Dict, Any

from phidata.asset.aws import AwsAsset, AwsAssetArgs
from phidata.infra.aws.resource.athena.query import AthenaQueryExecution
from phidata.utils.cli_console import print_info, print_error, print_warning
from phidata.utils.log import logger


class AthenaQueryArgs(AwsAssetArgs):
    # The SQL query statements to be executed.
    query_string: str
    # A unique case-sensitive string used to ensure the request to create the query
    # is idempotent (executes only once). If another StartQueryExecution request is received,
    # the same response is returned and another query is not created. If a parameter
    # has changed, for example, the QueryString , an error is returned.
    client_request_token: Optional[str] = None
    # QueryExecutionContext
    # The database within which the query executes.
    # The database must exist in the catalog.
    database: Optional[str] = None
    # The name of the data catalog used in the query execution.
    catalog: Optional[str] = None

    # ResultConfiguration
    # Specifies information about where and how to save the results of the query execution.
    # The location in Amazon S3 where your query results are stored,
    # such as s3://path/to/query/bucket/ . To run the query, you must specify the query
    # results location using one of the ways: either for individual queries using either
    # this setting (client-side), or in the workgroup, using WorkGroupConfiguration .
    # If none of them is set, Athena issues an error that no output location is provided.
    output_location: Optional[str] = None
    work_group: Optional[str] = None

    query_execution_id: Optional[str] = None
    wait_for_results: bool = True

    ## Resource management
    # If True, skip resource creation if active resources with the same name exist.
    use_cache: bool = True
    # If True, logs extra debug messages
    use_verbose_logs: bool = True


class AthenaQuery(AwsAsset):
    def __init__(
        self,
        # Name for this query.
        name: str,
        # The SQL query statements to be executed.
        query_string: str,
        # A unique case-sensitive string used to ensure the request to create the query
        # is idempotent (executes only once). If another StartQueryExecution request is received,
        # the same response is returned and another query is not created. If a parameter
        # has changed, for example, the QueryString , an error is returned.
        client_request_token: Optional[str] = None,
        # QueryExecutionContext
        # The database within which the query executes.
        # The database must exist in the catalog.
        database: Optional[str] = None,
        # The name of the data catalog used in the query execution.
        catalog: Optional[str] = None,
        # ResultConfiguration
        # Specifies information about where and how to save the results of the query execution.
        # The location in Amazon S3 where your query results are stored,
        # such as s3://path/to/query/bucket/ . To run the query, you must specify the query
        # results location using one of the ways: either for individual queries using either
        # this setting (client-side), or in the workgroup, using WorkGroupConfiguration .
        # If none of them is set, Athena issues an error that no output location is provided.
        output_location: Optional[str] = None,
        work_group: Optional[str] = None,
        query_execution_id: Optional[str] = None,
        wait_for_results: bool = True,
        ## Resource management
        # If True, skip resource creation if active resources with the same name exist.
        use_cache: bool = True,
        # If True, logs extra debug messages
        use_verbose_logs: bool = True,
        version: Optional[str] = None,
        enabled: bool = True,
    ) -> None:

        super().__init__()
        try:
            self.args: AthenaQueryArgs = AthenaQueryArgs(
                name=name,
                query_string=query_string,
                client_request_token=client_request_token,
                database=database,
                catalog=catalog,
                output_location=output_location,
                work_group=work_group,
                query_execution_id=query_execution_id,
                wait_for_results=wait_for_results,
                use_cache=use_cache,
                use_verbose_logs=use_verbose_logs,
                version=version,
                enabled=enabled,
            )
            self._athena_query_execution: Optional[AthenaQueryExecution] = None
        except Exception as e:
            print_error(f"Args for {self.__class__.__name__} are not valid")
            raise

    @property
    def athena_query_execution(self) -> AthenaQueryExecution:
        # use cached value if available
        if self._athena_query_execution:
            return self._athena_query_execution

        # Create AthenaQueryExecution object
        logger.debug("Init AthenaQueryExecution")
        try:
            query_execution_context = None
            if self.args.database or self.args.catalog:
                query_execution_context = {}
                if self.args.database:
                    query_execution_context["Database"] = self.args.database
                if self.args.catalog:
                    query_execution_context["Catalog"] = self.args.catalog

            result_configuration = None
            if self.args.output_location:
                result_configuration = {
                    "OutputLocation": self.args.output_location,
                }

            self._athena_query_execution = AthenaQueryExecution(
                name=self.name,
                query_string=self.args.query_string,
                client_request_token=self.args.client_request_token,
                query_execution_context=query_execution_context,
                result_configuration=result_configuration,
                work_group=self.args.work_group,
                query_execution_id=self.args.query_execution_id,
                use_cache=self.args.use_cache,
                use_verbose_logs=self.args.use_verbose_logs,
            )
            return self._athena_query_execution
        except Exception as e:
            print_error(f"Could not create AthenaQueryExecution")
            raise

    ######################################################
    ## Start AthenaQuery
    ######################################################

    def start_query(self) -> bool:

        # AthenaQuery not yet initialized
        if self.args is None:
            return False

        from phidata.infra.aws.api_client import AwsApiClient

        try:
            # aws_region, aws_profile loaded from local env if provided
            # by WorkspaceConfig
            aws_api_client = AwsApiClient(
                aws_region=self.aws_region,
                aws_profile=self.aws_profile,
            )
            athena_query_execution = self.athena_query_execution
            # logger.debug(f"Running: {self._athena_query_execution.dict()}")
            create_success = athena_query_execution.create(aws_api_client)
            return create_success
        except Exception as e:
            print_error("Could not start AthenaQuery, please try again")
            logger.exception(e)
        return False

    ######################################################
    ## Get results for AthenaQuery
    ######################################################

    def get_results(self) -> Any:

        # AthenaQuery not yet initialized
        if self.args is None:
            return False

        from phidata.infra.aws.api_client import AwsApiClient

        aws_api_client = AwsApiClient(
            aws_region=self.aws_region,
            aws_profile=self.aws_profile,
        )
        athena_query_execution = self.athena_query_execution
        active_athena_query_execution = athena_query_execution.read(aws_api_client)
        if active_athena_query_execution is None:
            print_error("No AthenaQuery available")
            return False

        results = athena_query_execution.get_results(aws_api_client)
        return results



from typing import Optional

from phidata.utils.log import logger
from phidata.task import PythonTask, PythonTaskArgs
from phidata.infra.aws.resource.glue.crawler import GlueCrawler


class StartGlueCrawlerArgs(PythonTaskArgs):
    crawler: GlueCrawler
    create_crawler: bool = True


class StartGlueCrawler(PythonTask):
    def __init__(
        self,
        crawler: GlueCrawler,
        create_crawler: bool = True,
        name: str = "start_glue_crawler",
        task_id: Optional[str] = None,
        dag_id: Optional[str] = None,
        version: Optional[str] = None,
        enabled: bool = True,
    ):
        super().__init__()
        try:
            self.args: StartGlueCrawlerArgs = StartGlueCrawlerArgs(
                crawler=crawler,
                create_crawler=create_crawler,
                name=name,
                task_id=task_id,
                dag_id=dag_id,
                version=version,
                enabled=enabled,
                entrypoint=start_glue_crawler,
            )
        except Exception as e:
            logger.error(f"Args for {self.__class__.__name__} are not valid")
            raise

    @property
    def crawler(self) -> GlueCrawler:
        return self.args.crawler

    @crawler.setter
    def crawler(self, crawler: GlueCrawler) -> None:
        if crawler is not None:
            self.args.crawler = crawler

    @property
    def create_crawler(self) -> bool:
        return self.args.create_crawler

    @create_crawler.setter
    def create_crawler(self, create_crawler: bool) -> None:
        if create_crawler is not None:
            self.args.create_crawler = create_crawler


def start_glue_crawler(**kwargs) -> bool:

    args: StartGlueCrawlerArgs = StartGlueCrawlerArgs(**kwargs)
    # logger.debug("StartGlueCrawlerArgs: {}".format(args))

    if args.create_crawler:
        create_success = args.crawler.create()
        if not create_success:
            logger.warning("Crawler could not be created.")

    return args.crawler.start_crawler()



from typing import Optional
from typing_extensions import Literal

import docker

from phidata.infra.base.api_client import InfraApiClient
from phidata.infra.docker.exceptions import DockerApiClientException
from phidata.utils.cli_console import print_error
from phidata.utils.log import logger


class DockerApiClient(InfraApiClient):
    """
    This class is a wrapper around the docker client to use with a DockerWorker
    Currently only supports local docker but can be modified to support a swarm cluster as well
    """

    def __init__(self, base_url: Optional[str] = None):

        super().__init__()
        # logger.debug(f"Creating DockerApiClient")
        # configuration
        self.base_url: Optional[str] = base_url

        # docker API client
        self._api_client: Optional[docker.DockerClient] = None
        logger.debug(f"**-+-** DockerApiClient created")

    def is_initialized(self) -> bool:
        if self.api_client is not None:
            return True
        return False

    def create_api_client(self) -> docker.DockerClient:
        """Create a docker.DockerApiClient"""

        logger.debug("Creating docker.DockerApiClient")
        api_client = None
        try:
            if self.base_url is None:
                api_client = docker.from_env()
            else:
                api_client = docker.DockerClient(base_url=self.base_url)
        except Exception as e:
            print_error("Could not create docker.DockerApiClient")
            print_error(e)
            exit(0)
        return api_client

    @property
    def api_client(self) -> docker.DockerClient:
        if self._api_client is None:
            self._api_client = self.create_api_client()
        return self._api_client



from collections import OrderedDict
from pathlib import Path
from typing import Optional, Dict, List, Union

from phidata.app.db import DbApp, DbAppArgs
from phidata.infra.aws.resource.ec2.volume import EbsVolume
from phidata.infra.docker.resource.container import DockerContainer
from phidata.infra.docker.resource.group import (
    DockerResourceGroup,
    DockerBuildContext,
)
from phidata.infra.docker.resource.network import DockerNetwork
from phidata.infra.k8s.create.apps.v1.deployment import CreateDeployment
from phidata.infra.k8s.create.common.port import CreatePort
from phidata.infra.k8s.create.core.v1.config_map import CreateConfigMap
from phidata.infra.k8s.create.core.v1.container import CreateContainer
from phidata.infra.k8s.create.core.v1.secret import CreateSecret
from phidata.infra.k8s.create.core.v1.service import CreateService, ServiceType
from phidata.infra.k8s.create.core.v1.volume import (
    CreateVolume,
    HostPathVolumeSource,
    AwsElasticBlockStoreVolumeSource,
    VolumeType,
)
from phidata.infra.k8s.create.group import CreateK8sResourceGroup
from phidata.infra.k8s.resource.group import (
    K8sResourceGroup,
    K8sBuildContext,
)
from phidata.infra.k8s.enums.image_pull_policy import ImagePullPolicy
from phidata.infra.k8s.enums.restart_policy import RestartPolicy
from phidata.utils.cli_console import print_error, print_info, print_warning
from phidata.utils.common import (
    get_image_str,
    get_default_container_name,
    get_default_configmap_name,
    get_default_secret_name,
    get_default_service_name,
    get_default_volume_name,
    get_default_deploy_name,
    get_default_pod_name,
)
from phidata.utils.enums import ExtendedEnum
from phidata.utils.log import logger


class PostgresVolumeType(ExtendedEnum):
    EMPTY_DIR = "EMPTY_DIR"
    HOST_PATH = "HOST_PATH"
    PERSISTENT_VOLUME = "PERSISTENT_VOLUME"
    AWS_EBS = "AWS_EBS"


class PostgresDbArgs(DbAppArgs):
    name: str = "postgres"
    version: str = "1"
    enabled: bool = True

    # Image args
    image_name: str = "postgres"
    image_tag: str = "14"
    entrypoint: Optional[Union[str, List]] = None
    command: Optional[Union[str, List]] = None

    # Configure postgres
    # Provide POSTGRES_USER as db_user or POSTGRES_USER in secrets_file
    db_user: Optional[str] = None
    # Provide POSTGRES_PASSWORD as db_password or POSTGRES_PASSWORD in secrets_file
    db_password: Optional[str] = None
    # Provide POSTGRES_DB as db_schema or POSTGRES_DB insecretsv_file
    db_schema: Optional[str] = None
    db_driver: str = "postgresql"
    pgdata: Optional[str] = "/var/lib/postgresql/data/pgdata"
    postgres_initdb_args: Optional[str] = None
    postgres_initdb_waldir: Optional[str] = None
    postgres_host_auth_method: Optional[str] = None
    postgres_password_file: Optional[str] = None
    postgres_user_file: Optional[str] = None
    postgres_db_file: Optional[str] = None
    postgres_initdb_args_file: Optional[str] = None

    # Configure the volume
    create_volume: bool = True
    volume_type: PostgresVolumeType = PostgresVolumeType.EMPTY_DIR
    volume_name: Optional[str] = None
    # Container path to mount the postgres volume
    # should be the parent directory of pgdata
    volume_container_path: str = "/var/lib/postgresql/data"
    # Host path to mount the postgres volume
    # If volume_type = PostgresVolumeType.HOST_PATH
    volume_host_path: Optional[Path] = None
    # EbsVolume if volume_type = PostgresVolumeType.AWS_EBS
    ebs_volume: Optional[EbsVolume] = None
    # EbsVolume region is used to determine the ebs_volume_id
    # and add topology region selectors
    ebs_volume_region: Optional[str] = None
    # Provide Ebs Volume-id manually
    ebs_volume_id: Optional[str] = None
    # Add topology az selectors
    ebs_volume_az: Optional[str] = None
    # Add NodeSelectors to Pods, so they are scheduled in the same
    # region and zone as the ebs_volume
    schedule_pods_in_ebs_topology: bool = True

    # Configure the container
    container_name: Optional[str] = None
    image_pull_policy: ImagePullPolicy = ImagePullPolicy.IF_NOT_PRESENT
    container_port: int = 5432
    # Only used by the K8sContainer
    container_port_name: str = "pg"
    # Only used by the DockerContainer
    container_host_port: int = 5432
    container_detach: bool = True
    container_auto_remove: bool = True
    container_remove: bool = True

    # Add env variables to container env
    env: Optional[Dict[str, str]] = None
    # Read env variables from a file in yaml format
    env_file: Optional[Path] = None
    # Configure the ConfigMap used for env variables that are not Secret
    config_map_name: Optional[str] = None
    # Configure the Secret used for env variables that are Secret
    secret_name: Optional[str] = None
    # Read secrets from a file in yaml format
    secrets_file: Optional[Path] = None

    # Configure the deployment
    deploy_name: Optional[str] = None
    pod_name: Optional[str] = None
    replicas: int = 1
    pod_node_selector: Optional[Dict[str, str]] = None
    restart_policy: RestartPolicy = RestartPolicy.ALWAYS
    termination_grace_period_seconds: Optional[int] = None

    # Configure the service
    service_name: Optional[str] = None
    service_type: Optional[ServiceType] = None
    # The port that will be exposed by the service.
    service_port: int = 5432
    # The node_port that will be exposed by the service if service_type = ServiceType.NODE_PORT
    node_port: Optional[int] = None
    # The target_port is the port to access on the pods targeted by the service.
    # It can be the port number or port name on the pod.
    target_port: Optional[Union[str, int]] = None


class PostgresDb(DbApp):
    def __init__(
        self,
        name: str = "postgres",
        version: str = "1",
        enabled: bool = True,
        # Image args,
        image_name: str = "postgres",
        image_tag: str = "14",
        entrypoint: Optional[Union[str, List]] = None,
        command: Optional[Union[str, List]] = None,
        # Configure postgres,
        # Provide POSTGRES_USER as db_user or POSTGRES_USER in secrets_file,
        db_user: Optional[str] = None,
        # Provide POSTGRES_PASSWORD as db_password or POSTGRES_PASSWORD in secrets_file,
        db_password: Optional[str] = None,
        # Provide POSTGRES_DB as db_schema or POSTGRES_DB in secrets_file,
        db_schema: Optional[str] = None,
        db_driver: str = "postgresql",
        pgdata: Optional[str] = "/var/lib/postgresql/data/pgdata",
        postgres_initdb_args: Optional[str] = None,
        postgres_initdb_waldir: Optional[str] = None,
        postgres_host_auth_method: Optional[str] = None,
        postgres_password_file: Optional[str] = None,
        postgres_user_file: Optional[str] = None,
        postgres_db_file: Optional[str] = None,
        postgres_initdb_args_file: Optional[str] = None,
        # Configure the volume,
        create_volume: bool = True,
        volume_type: PostgresVolumeType = PostgresVolumeType.EMPTY_DIR,
        volume_name: Optional[str] = None,
        # Container path to mount the postgres volume,
        # should be the parent directory of pgdata,
        volume_container_path: str = "/var/lib/postgresql/data",
        # Host path to mount the postgres volume,
        # If volume_type = PostgresVolumeType.HOST_PATH,
        volume_host_path: Optional[Path] = None,
        # EbsVolume if volume_type = PostgresVolumeType.AWS_EBS
        ebs_volume: Optional[EbsVolume] = None,
        ebs_volume_region: Optional[str] = None,
        # Provide Ebs Volume-id manually
        ebs_volume_id: Optional[str] = None,
        # Add topology az selectors
        ebs_volume_az: Optional[str] = None,
        schedule_pods_in_ebs_topology: bool = True,
        # Configure the container,
        container_name: Optional[str] = None,
        image_pull_policy: ImagePullPolicy = ImagePullPolicy.IF_NOT_PRESENT,
        container_port: int = 5432,
        # Only used by the K8sContainer,
        container_port_name: str = "pg",
        # Only used by the DockerContainer,
        container_host_port: int = 5432,
        container_detach: bool = True,
        container_auto_remove: bool = True,
        container_remove: bool = True,
        # Add env variables to container env,
        env: Optional[Dict[str, str]] = None,
        # Read env variables from a file in yaml format,
        env_file: Optional[Path] = None,
        # Configure the ConfigMap used for env variables that are not Secret,
        config_map_name: Optional[str] = None,
        # Configure the Secret used for env variables that are Secret,
        secret_name: Optional[str] = None,
        # Read secrets from a file in yaml format,
        secrets_file: Optional[Path] = None,
        # Configure the deployment,
        deploy_name: Optional[str] = None,
        pod_name: Optional[str] = None,
        replicas: int = 1,
        pod_node_selector: Optional[Dict[str, str]] = None,
        restart_policy: RestartPolicy = RestartPolicy.ALWAYS,
        termination_grace_period_seconds: Optional[int] = None,
        # Configure the service,
        service_name: Optional[str] = None,
        service_type: Optional[ServiceType] = None,
        # The port that will be exposed by the service.,
        service_port: int = 5432,
        # The node_port that will be exposed by the service if service_type = ServiceType.NODE_PORT,
        node_port: Optional[int] = None,
        # The target_port is the port to access on the pods targeted by the service.,
        # It can be the port number or port name on the pod.,
        target_port: Optional[Union[str, int]] = None,
        # Additional args
        # If True, skip resource creation if active resources with the same name exist.
        use_cache: bool = True,
        # If True, log extra debug messages
        use_verbose_logs: bool = False,
    ):
        super().__init__()
        try:
            self.args: PostgresDbArgs = PostgresDbArgs(
                name=name,
                version=version,
                enabled=enabled,
                image_name=image_name,
                image_tag=image_tag,
                entrypoint=entrypoint,
                command=command,
                db_user=db_user,
                db_password=db_password,
                db_schema=db_schema,
                db_driver=db_driver,
                pgdata=pgdata,
                postgres_initdb_args=postgres_initdb_args,
                postgres_initdb_waldir=postgres_initdb_waldir,
                postgres_host_auth_method=postgres_host_auth_method,
                postgres_password_file=postgres_password_file,
                postgres_user_file=postgres_user_file,
                postgres_db_file=postgres_db_file,
                postgres_initdb_args_file=postgres_initdb_args_file,
                create_volume=create_volume,
                volume_type=volume_type,
                volume_name=volume_name,
                volume_container_path=volume_container_path,
                volume_host_path=volume_host_path,
                ebs_volume=ebs_volume,
                ebs_volume_region=ebs_volume_region,
                ebs_volume_id=ebs_volume_id,
                ebs_volume_az=ebs_volume_az,
                schedule_pods_in_ebs_topology=schedule_pods_in_ebs_topology,
                container_name=container_name,
                image_pull_policy=image_pull_policy,
                container_port=container_port,
                container_port_name=container_port_name,
                container_host_port=container_host_port,
                container_detach=container_detach,
                container_auto_remove=container_auto_remove,
                container_remove=container_remove,
                env=env,
                env_file=env_file,
                config_map_name=config_map_name,
                secret_name=secret_name,
                secrets_file=secrets_file,
                deploy_name=deploy_name,
                pod_name=pod_name,
                replicas=replicas,
                pod_node_selector=pod_node_selector,
                restart_policy=restart_policy,
                termination_grace_period_seconds=termination_grace_period_seconds,
                service_name=service_name,
                service_type=service_type,
                service_port=service_port,
                node_port=node_port,
                target_port=target_port,
                use_cache=use_cache,
                use_verbose_logs=use_verbose_logs,
            )
        except Exception:
            logger.error(f"Args for {self.__class__.__name__} are not valid")
            raise

    def get_container_name(self) -> str:
        return self.args.container_name or get_default_container_name(self.args.name)

    def get_service_name(self) -> str:
        return self.args.service_name or get_default_service_name(self.args.name)

    def get_service_port(self) -> int:
        return self.args.service_port

    def get_env_data_from_file(self) -> Optional[Dict[str, str]]:
        import yaml

        env_file_path = self.args.env_file
        if (
            env_file_path is not None
            and env_file_path.exists()
            and env_file_path.is_file()
        ):
            # logger.debug(f"Reading {env_file_path}")
            env_data_from_file = yaml.safe_load(env_file_path.read_text())
            if env_data_from_file is not None and isinstance(env_data_from_file, dict):
                return env_data_from_file
            else:
                print_error(f"Invalid env_file: {env_file_path}")
        return None

    def get_secret_data_from_file(self) -> Optional[Dict[str, str]]:
        import yaml

        secrets_file_path = self.args.secrets_file
        if (
            secrets_file_path is not None
            and secrets_file_path.exists()
            and secrets_file_path.is_file()
        ):
            # logger.debug(f"Reading {secrets_file_path}")
            secret_data_from_file = yaml.safe_load(secrets_file_path.read_text())
            if secret_data_from_file is not None and isinstance(
                secret_data_from_file, dict
            ):
                return secret_data_from_file
            else:
                print_error(f"Invalid secrets_file: {secrets_file_path}")
        return None

    def get_db_user(self) -> Optional[str]:
        db_user_var: Optional[str] = self.args.db_user if self.args else None
        if db_user_var is None and self.args.secrets_file is not None:
            # read from secrets_file
            # logger.debug(f"Reading POSTGRES_USER from secrets_file")
            secret_data_from_file = self.get_secret_data_from_file()
            if secret_data_from_file is not None:
                db_user_var = secret_data_from_file.get("POSTGRES_USER", db_user_var)
        return db_user_var

    def get_db_password(self) -> Optional[str]:
        db_password_var: Optional[str] = self.args.db_password if self.args else None
        if db_password_var is None and self.args.secrets_file is not None:
            # read from secrets_file
            # logger.debug(f"Reading POSTGRES_PASSWORD from secrets_file")
            secret_data_from_file = self.get_secret_data_from_file()
            if secret_data_from_file is not None:
                db_password_var = secret_data_from_file.get(
                    "POSTGRES_PASSWORD", db_password_var
                )
        return db_password_var

    def get_db_schema(self) -> Optional[str]:
        db_schema_var: Optional[str] = self.args.db_schema if self.args else None
        if db_schema_var is None and self.args.secrets_file is not None:
            # read from secrets_file
            # logger.debug(f"Reading POSTGRES_DB from secrets_file")
            secret_data_from_file = self.get_secret_data_from_file()
            if secret_data_from_file is not None:
                db_schema_var = secret_data_from_file.get("POSTGRES_DB", db_schema_var)
        return db_schema_var

    def get_db_driver(self) -> Optional[str]:
        return self.args.db_driver if self.args else "postgresql"

    def get_db_host_local(self) -> Optional[str]:
        return "localhost"

    def get_db_port_local(self) -> Optional[int]:
        return self.args.container_host_port if self.args else None

    def get_db_host_docker(self) -> Optional[str]:
        return self.get_container_name()

    def get_db_port_docker(self) -> Optional[int]:
        return self.args.container_port if self.args else None

    def get_db_host_k8s(self) -> Optional[str]:
        return self.get_service_name()

    def get_db_port_k8s(self) -> Optional[int]:
        return self.get_service_port()

    def get_db_connection_url_local(self) -> Optional[str]:
        user = self.get_db_user()
        password = self.get_db_password()
        schema = self.get_db_schema()
        driver = self.get_db_driver()
        host = self.get_db_host_local()
        port = self.get_db_port_local()
        return f"{driver}://{user}:{password}@{host}:{port}/{schema}"

    def get_db_connection_url_docker(self) -> Optional[str]:
        user = self.get_db_user()
        password = self.get_db_password()
        schema = self.get_db_schema()
        driver = self.get_db_driver()
        host = self.get_db_host_docker()
        port = self.get_db_port_docker()
        return f"{driver}://{user}:{password}@{host}:{port}/{schema}"

    def get_db_connection_url_k8s(self) -> Optional[str]:
        user = self.get_db_user()
        password = self.get_db_password()
        schema = self.get_db_schema()
        driver = self.get_db_driver()
        host = self.get_db_host_k8s()
        port = self.get_db_port_k8s()
        return f"{driver}://{user}:{password}@{host}:{port}/{schema}"

    ######################################################
    ## Docker Resources
    ######################################################

    def get_docker_rg(
        self, docker_build_context: DockerBuildContext
    ) -> Optional[DockerResourceGroup]:

        app_name = self.args.name
        logger.debug(f"Building {app_name} DockerResourceGroup")

        # Container Environment
        container_env: Dict[str, str] = {}
        if self.args.db_user:
            container_env["POSTGRES_USER"] = self.args.db_user
        if self.args.db_password:
            container_env["POSTGRES_PASSWORD"] = self.args.db_password
        if self.args.db_schema:
            container_env["POSTGRES_DB"] = self.args.db_schema
        if self.args.pgdata:
            container_env["PGDATA"] = self.args.pgdata
        if self.args.postgres_initdb_args:
            container_env["POSTGRES_INITDB_ARGS"] = self.args.postgres_initdb_args
        if self.args.postgres_initdb_waldir:
            container_env["POSTGRES_INITDB_WALDIR"] = self.args.postgres_initdb_waldir
        if self.args.postgres_host_auth_method:
            container_env[
                "POSTGRES_HOST_AUTH_METHOD"
            ] = self.args.postgres_host_auth_method
        if self.args.postgres_password_file:
            container_env["POSTGRES_PASSWORD_FILE"] = self.args.postgres_password_file
        if self.args.postgres_user_file:
            container_env["POSTGRES_USER_FILE"] = self.args.postgres_user_file
        if self.args.postgres_db_file:
            container_env["POSTGRES_DB_FILE"] = self.args.postgres_db_file
        if self.args.postgres_initdb_args_file:
            container_env[
                "POSTGRES_INITDB_ARGS_FILE"
            ] = self.args.postgres_initdb_args_file

        # Update the container env using env_file
        env_data_from_file = self.get_env_data_from_file()
        if env_data_from_file is not None:
            container_env.update(env_data_from_file)

        # Update the container env using secrets_file
        secret_data_from_file = self.get_secret_data_from_file()
        if secret_data_from_file is not None:
            container_env.update(secret_data_from_file)

        # Update the container env with user provided env
        if self.args.env is not None and isinstance(self.args.env, dict):
            container_env.update(self.args.env)

        # Container Volumes
        container_volumes = {}
        if self.args.create_volume:
            if self.args.volume_type == PostgresVolumeType.EMPTY_DIR:
                volume_name = self.args.volume_name or get_default_volume_name(app_name)
                container_volumes[volume_name] = {
                    "bind": self.args.volume_container_path,
                    "mode": "rw",
                }
            elif self.args.volume_type == PostgresVolumeType.HOST_PATH:
                if self.args.volume_host_path is not None:
                    volume_host_path_str = str(self.args.volume_host_path)
                    container_volumes[volume_host_path_str] = {
                        "bind": self.args.volume_container_path,
                        "mode": "rw",
                    }
                else:
                    print_error("PostgresDb: volume_host_path not provided")
                    return None
            else:
                print_error(f"{self.args.volume_type.value} not supported")
                return None

        # Container Ports
        container_ports: Dict[str, int] = {
            str(self.args.container_port): self.args.container_host_port,
        }

        # Create the container
        docker_container = DockerContainer(
            name=self.get_container_name(),
            image=get_image_str(self.args.image_name, self.args.image_tag),
            entrypoint=self.args.entrypoint,
            command=self.args.command,
            detach=self.args.container_detach,
            auto_remove=self.args.container_auto_remove,
            remove=self.args.container_remove,
            stdin_open=True,
            tty=True,
            environment=container_env,
            network=docker_build_context.network,
            ports=container_ports,
            volumes=container_volumes,
            use_cache=self.args.use_cache,
            use_verbose_logs=self.args.use_verbose_logs,
        )

        docker_rg = DockerResourceGroup(
            name=app_name,
            enabled=self.args.enabled,
            network=DockerNetwork(name=docker_build_context.network),
            containers=[docker_container],
        )
        return docker_rg

    def init_docker_resource_groups(
        self, docker_build_context: DockerBuildContext
    ) -> None:
        docker_rg = self.get_docker_rg(docker_build_context)
        if docker_rg is not None:
            if self.docker_resource_groups is None:
                self.docker_resource_groups = OrderedDict()
            self.docker_resource_groups[docker_rg.name] = docker_rg

    ######################################################
    ## K8s Resources
    ######################################################

    def get_k8s_rg(
        self, k8s_build_context: K8sBuildContext
    ) -> Optional[K8sResourceGroup]:

        app_name = self.args.name
        logger.debug(f"Building {app_name} K8sResourceGroup")

        # Container Environment
        container_env: Dict[str, str] = {}
        if self.args.db_user:
            container_env["POSTGRES_USER"] = self.args.db_user
        if self.args.db_password:
            container_env["POSTGRES_PASSWORD"] = self.args.db_password
        if self.args.db_schema:
            container_env["POSTGRES_DB"] = self.args.db_schema
        if self.args.pgdata:
            container_env["PGDATA"] = self.args.pgdata
        if self.args.postgres_initdb_args:
            container_env["POSTGRES_INITDB_ARGS"] = self.args.postgres_initdb_args
        if self.args.postgres_initdb_waldir:
            container_env["POSTGRES_INITDB_WALDIR"] = self.args.postgres_initdb_waldir
        if self.args.postgres_host_auth_method:
            container_env[
                "POSTGRES_HOST_AUTH_METHOD"
            ] = self.args.postgres_host_auth_method
        if self.args.postgres_password_file:
            container_env["POSTGRES_PASSWORD_FILE"] = self.args.postgres_password_file
        if self.args.postgres_user_file:
            container_env["POSTGRES_USER_FILE"] = self.args.postgres_user_file
        if self.args.postgres_db_file:
            container_env["POSTGRES_DB_FILE"] = self.args.postgres_db_file
        if self.args.postgres_initdb_args_file:
            container_env[
                "POSTGRES_INITDB_ARGS_FILE"
            ] = self.args.postgres_initdb_args_file

        # Update the container env using env_file
        env_data_from_file = self.get_env_data_from_file()
        if env_data_from_file is not None:
            container_env.update(env_data_from_file)
        # Update the container env with user provided env
        if self.args.env is not None and isinstance(self.args.env, dict):
            container_env.update(self.args.env)
        # Create a ConfigMap to set the container env variables which are not Secret
        container_env_cm = CreateConfigMap(
            cm_name=self.args.config_map_name or get_default_configmap_name(app_name),
            app_name=app_name,
            data=container_env,
        )

        # Create a Secret to set the container env variables which are Secret
        container_env_secret: Optional[CreateSecret] = None
        secret_data_from_file = self.get_secret_data_from_file()
        if secret_data_from_file is not None:
            container_env_secret = CreateSecret(
                secret_name=self.args.secret_name or get_default_secret_name(app_name),
                app_name=app_name,
                string_data=secret_data_from_file,
            )

        # Container Volumes
        container_volumes = []
        # Add NodeSelectors to Pods in case we create az sensitive volumes
        pod_node_selector: Optional[Dict[str, str]] = self.args.pod_node_selector
        if self.args.create_volume:
            volume_name = self.args.volume_name or get_default_volume_name(app_name)
            if self.args.volume_type == PostgresVolumeType.EMPTY_DIR:
                pg_volume = CreateVolume(
                    volume_name=volume_name,
                    app_name=app_name,
                    mount_path=self.args.volume_container_path,
                    volume_type=VolumeType.EMPTY_DIR,
                )
                container_volumes.append(pg_volume)
            elif self.args.volume_type == PostgresVolumeType.HOST_PATH:
                if self.args.volume_host_path is not None:
                    volume_host_path_str = str(self.args.volume_host_path)
                    pg_volume = CreateVolume(
                        volume_name=volume_name,
                        app_name=app_name,
                        mount_path=self.args.volume_container_path,
                        volume_type=VolumeType.HOST_PATH,
                        host_path=HostPathVolumeSource(
                            path=volume_host_path_str,
                        ),
                    )
                    container_volumes.append(pg_volume)
                else:
                    print_error("PostgresDb: volume_host_path not provided")
                    return None
            elif self.args.volume_type == PostgresVolumeType.AWS_EBS:
                if (
                    self.args.ebs_volume_id is not None
                    or self.args.ebs_volume is not None
                ):
                    ebs_volume_id = self.args.ebs_volume_id
                    if ebs_volume_id is None and self.args.ebs_volume is not None:
                        ebs_volume_id = self.args.ebs_volume.get_volume_id(
                            aws_region=self.args.ebs_volume_region
                        )
                    logger.debug(f"ebs_volume_id: {ebs_volume_id}")

                    pg_volume = CreateVolume(
                        volume_name=volume_name,
                        app_name=app_name,
                        mount_path=self.args.volume_container_path,
                        volume_type=VolumeType.AWS_EBS,
                        aws_ebs=AwsElasticBlockStoreVolumeSource(
                            volume_id=ebs_volume_id,
                        ),
                    )
                    container_volumes.append(pg_volume)
                    if self.args.schedule_pods_in_ebs_topology:
                        if pod_node_selector is None:
                            pod_node_selector = {}

                        # Add NodeSelectors to Pods, so they are scheduled in the same
                        # region and zone as the ebs_volume
                        # https://kubernetes.io/docs/reference/labels-annotations-taints/#topologykubernetesiozone
                        ebs_region = self.args.ebs_volume_region
                        if ebs_region is not None:
                            pod_node_selector[
                                "topology.kubernetes.io/region"
                            ] = ebs_region

                        ebs_az = self.args.ebs_volume_az
                        if ebs_az is None and self.args.ebs_volume is not None:
                            ebs_az = self.args.ebs_volume.availability_zone
                        if ebs_az is not None:
                            pod_node_selector["topology.kubernetes.io/zone"] = ebs_az
                else:
                    print_error("PostgresDb: ebs_volume not provided")
                    return None
            else:
                print_error(f"{self.args.volume_type.value} not supported")
                return None

        # Create the ports to open
        container_port = CreatePort(
            name=self.args.container_port_name,
            container_port=self.args.container_port,
            service_port=self.args.service_port,
            target_port=self.args.target_port or self.args.container_port_name,
        )

        # If ServiceType == NODE_PORT then validate self.args.node_port is available
        if self.args.service_type == ServiceType.NODE_PORT:
            if (
                self.args.node_port is None
                or self.args.node_port < 30000
                or self.args.node_port > 32767
            ):
                print_error(f"NodePort: {self.args.node_port} invalid")
                print_error(f"Skipping this service")
                return None
            else:
                container_port.node_port = self.args.node_port
        # If ServiceType == LOAD_BALANCER then validate self.args.node_port only IF available
        elif self.args.service_type == ServiceType.LOAD_BALANCER:
            if self.args.node_port is not None:
                if self.args.node_port < 30000 or self.args.node_port > 32767:
                    print_error(f"NodePort: {self.args.node_port} invalid")
                    print_error(f"Skipping this service")
                    return None
                else:
                    container_port.node_port = self.args.node_port
        # else validate self.args.node_port is NOT available
        elif self.args.node_port is not None:
            print_warning(
                f"NodePort: {self.args.node_port} provided without specifying ServiceType as NODE_PORT or LOAD_BALANCER"
            )
            print_warning("NodePort value will be ignored")
            self.args.node_port = None

        # Create the container
        k8s_container = CreateContainer(
            container_name=self.get_container_name(),
            app_name=app_name,
            image_name=self.args.image_name,
            image_tag=self.args.image_tag,
            # Equivalent to docker images CMD
            args=[self.args.command]
            if isinstance(self.args.command, str)
            else self.args.command,
            # Equivalent to docker images ENTRYPOINT
            command=self.args.entrypoint,
            image_pull_policy=self.args.image_pull_policy,
            envs_from_configmap=[container_env_cm.cm_name],
            envs_from_secret=[container_env_secret.secret_name]
            if container_env_secret
            else None,
            ports=[container_port],
            volumes=container_volumes,
            labels=k8s_build_context.labels,
        )

        # Create the deployment
        k8s_deployment = CreateDeployment(
            deploy_name=self.args.deploy_name or get_default_deploy_name(app_name),
            pod_name=self.args.pod_name or get_default_pod_name(app_name),
            app_name=app_name,
            namespace=k8s_build_context.namespace,
            service_account_name=k8s_build_context.service_account_name,
            replicas=self.args.replicas,
            containers=[k8s_container],
            pod_node_selector=pod_node_selector,
            restart_policy=self.args.restart_policy,
            termination_grace_period_seconds=self.args.termination_grace_period_seconds,
            volumes=container_volumes,
            labels=k8s_build_context.labels,
        )

        # Create the service
        k8s_service = CreateService(
            service_name=self.get_service_name(),
            app_name=app_name,
            namespace=k8s_build_context.namespace,
            service_account_name=k8s_build_context.service_account_name,
            service_type=self.args.service_type,
            deployment=k8s_deployment,
            ports=[container_port],
            labels=k8s_build_context.labels,
        )

        # Create the K8sResourceGroup
        k8s_resource_group = CreateK8sResourceGroup(
            name=app_name,
            enabled=self.args.enabled,
            config_maps=[container_env_cm],
            secrets=[container_env_secret] if container_env_secret else None,
            services=[k8s_service],
            deployments=[k8s_deployment],
        )

        return k8s_resource_group.create()

    def init_k8s_resource_groups(self, k8s_build_context: K8sBuildContext) -> None:
        k8s_rg = self.get_k8s_rg(k8s_build_context)
        if k8s_rg is not None:
            if self.k8s_resource_groups is None:
                self.k8s_resource_groups = OrderedDict()
            self.k8s_resource_groups[k8s_rg.name] = k8s_rg



from typing import List, Optional

from kubernetes.client import CoreV1Api
from kubernetes.client.models.v1_pod import V1Pod
from kubernetes.client.models.v1_pod_list import V1PodList

from phidata.infra.k8s.api_client import K8sApiClient
from phidata.infra.k8s.resource.base import K8sResource
from phidata.utils.log import logger


class Pod(K8sResource):
    """
    There are no attributes in the Pod model because we don't create Pods manually.
    This class exists only to read from the K8s cluster.

    References:
        * Doc: https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#pod-v1-core
        * Type: https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_pod.py
    """

    resource_type: str = "Pod"

    @staticmethod
    def get_from_cluster(
        k8s_client: K8sApiClient, namespace: Optional[str] = None, **kwargs: str
    ) -> Optional[List[V1Pod]]:

        core_v1_api: CoreV1Api = k8s_client.core_v1_api
        pod_list: Optional[V1PodList] = None
        if namespace:
            # logger.debug(f"Getting Pods for ns: {namespace}")
            pod_list = core_v1_api.list_namespaced_pod(namespace=namespace)
        else:
            # logger.debug("Getting SA for all namespaces")
            pod_list = core_v1_api.list_pod_for_all_namespaces()

        pods: Optional[List[V1Pod]] = None
        if pod_list:
            pods = pod_list.items
        # logger.debug(f"pods: {pods}")
        # logger.debug(f"pods type: {type(pods)}")
        return pods

    def _read(self, k8s_client: K8sApiClient) -> Optional[V1Pod]:
        """Returns the "Active" Deployment from the cluster"""

        namespace = self.get_namespace()
        active_resource: Optional[V1Pod] = None
        active_resources: Optional[List[V1Pod]] = self.get_from_cluster(
            k8s_client=k8s_client,
            namespace=namespace,
        )
        # logger.debug(f"Active Resources: {active_resources}")
        if active_resources is None or len(active_resources) == 0:
            return None

        resource_name = self.get_resource_name()
        logger.debug("resource_name: {}".format(resource_name))
        for resource in active_resources:
            logger.debug(f"Checking {resource.metadata.name}")
            pod_name = ""
            try:
                pod_name = resource.metadata.name
            except Exception as e:
                logger.error(f"Cannot read pod name: {e}")
                continue
            if resource_name in pod_name:
                self.active_resource = resource
                logger.debug(f"Found active {resource_name}")
                break

        return self.active_resource



from typing import Optional, Any, Dict

from phidata.base import PhidataBase, PhidataBaseArgs
from phidata.utils.context import get_run_date
from phidata.utils.env_var import validate_env_vars
from phidata.utils.log import logger
from phidata.types.context import PathContext, RunContext, AirflowContext


class TaskArgs(PhidataBaseArgs):
    # airflow task_id, use name if not provided
    task_id: Optional[str] = None
    # airflow dag_id for this task, set by the workflow
    dag_id: Optional[str] = None

    # run_context provided by workflow
    run_context: Optional[RunContext] = None
    # path_context provided by workflow
    path_context: Optional[PathContext] = None
    # airflow_context provided by workflow
    airflow_context: Optional[AirflowContext] = None

    @property
    def run_date(self) -> str:
        return get_run_date(self.run_context, self.airflow_context)

    @classmethod
    def from_kwargs(cls, kwargs: Optional[Dict] = None):
        if kwargs is None or not isinstance(kwargs, dict):
            return cls()
        # logger.info(f"Loading {cls.__name__} using kwargs")
        args_object = cls(**kwargs)
        validate_airflow_env = kwargs.get("validate_airflow_env", None)
        if validate_env_vars(validate_airflow_env):
            # logger.info("Creating airflow_context")
            airflow_context = AirflowContext(**kwargs)
            args_object.airflow_context = airflow_context
        return args_object


class Task(PhidataBase):
    """Base Class for all Tasks"""

    def __init__(
        self,
        *,
        name: Optional[str] = None,
        task_id: Optional[str] = None,
        dag_id: Optional[str] = None,
        version: Optional[str] = None,
        enabled: bool = True,
        run_context: Optional[RunContext] = None,
        path_context: Optional[PathContext] = None,
        airflow_context: Optional[AirflowContext] = None,
    ) -> None:
        super().__init__()
        self.args: Optional[TaskArgs] = None
        if name is not None and task_id is not None:
            self.args = TaskArgs(
                name=name,
                task_id=task_id,
                dag_id=dag_id,
                version=version,
                enabled=enabled,
                run_context=run_context,
                path_context=path_context,
                airflow_context=airflow_context,
            )

    @property
    def task_id(self) -> str:
        return self.args.task_id if (self.args and self.args.task_id) else self.name

    @task_id.setter
    def task_id(self, task_id: str) -> None:
        if self.args is not None and task_id is not None:
            self.args.task_id = task_id

    @property
    def dag_id(self) -> str:
        return self.args.dag_id if (self.args and self.args.dag_id) else self.name

    @dag_id.setter
    def dag_id(self, dag_id: str) -> None:
        if self.args is not None and dag_id is not None:
            self.args.dag_id = dag_id

    @property
    def run_context(self) -> Optional[RunContext]:
        return self.args.run_context if self.args else None

    @run_context.setter
    def run_context(self, run_context: RunContext) -> None:
        if self.args is not None and run_context is not None:
            self.args.run_context = run_context

    @property
    def path_context(self) -> Optional[PathContext]:
        return self.args.path_context if self.args else None

    @path_context.setter
    def path_context(self, path_context: PathContext) -> None:
        if self.args is not None and path_context is not None:
            self.args.path_context = path_context

    @property
    def airflow_context(self) -> Optional[AirflowContext]:
        return self.args.airflow_context if self.args else None

    @airflow_context.setter
    def airflow_context(self, airflow_context: AirflowContext) -> None:
        if self.args is not None and airflow_context is not None:
            self.args.airflow_context = airflow_context

    ######################################################
    ## Build task
    ######################################################

    def build(self) -> bool:
        logger.debug(f"@build not defined for {self.__class__.__name__}")
        return False

    ######################################################
    ## Run task
    ######################################################

    def run_in_local_env(self) -> bool:
        """
        Runs a task in the local environment.

        Returns:
            run_status (bool): True if the run was successful
        """
        logger.debug(f"@run_in_local_env not defined for {self.__class__.__name__}")
        return False

    def run_in_docker_container(
        self,
        active_container: Any,
        docker_env: Optional[Dict[str, str]] = None,
    ) -> bool:
        """
        Runs a task in a docker container.

        Args:
            active_container:
            docker_env:

        Returns:
            run_status (bool): True if the run was successful

        Notes:
            * This function runs in the local environment where phi wf is called from.
            But executes `airflow` commands in the docker container to run the task
        """
        logger.debug(
            f"@run_in_docker_container not defined for {self.__class__.__name__}"
        )
        return False

    def run_in_k8s_container(
        self,
        pod: Any,
        k8s_api_client: Any,
        container_name: Optional[str] = None,
        k8s_env: Optional[Dict[str, str]] = None,
    ) -> bool:
        logger.debug(f"@run_in_k8s_container not defined for {self.__class__.__name__}")
        return False

    ######################################################
    ## Airflow functions
    ######################################################

    def add_airflow_task(
        self, dag: Any, task_group: Optional[Any] = None
    ) -> Optional[Any]:
        """
        Adds the airflow task to the dag and task_group.
        Returns the Operator

        Returns:
            Any subclass of BaseOperator
        """
        logger.debug(f"@add_airflow_task not defined for {self.__class__.__name__}")
        return None



from typing import Optional

from pydantic import BaseModel

from phidata.infra.k8s.enums.volume_type import VolumeType
from phidata.infra.k8s.resource.core.v1.volume import (
    Volume,
    AwsElasticBlockStoreVolumeSource,
    PersistentVolumeClaimVolumeSource,
    GcePersistentDiskVolumeSource,
    SecretVolumeSource,
    EmptyDirVolumeSource,
    ConfigMapVolumeSource,
    GitRepoVolumeSource,
    HostPathVolumeSource,
)
from phidata.utils.cli_console import print_error, print_info
from phidata.utils.log import logger


class CreateVolume(BaseModel):
    volume_name: str
    app_name: str
    mount_path: str
    volume_type: VolumeType
    aws_ebs: Optional[AwsElasticBlockStoreVolumeSource] = None
    config_map: Optional[ConfigMapVolumeSource] = None
    empty_dir: Optional[EmptyDirVolumeSource] = None
    gce_persistent_disk: Optional[GcePersistentDiskVolumeSource] = None
    git_repo: Optional[GitRepoVolumeSource] = None
    host_path: Optional[HostPathVolumeSource] = None
    pvc: Optional[PersistentVolumeClaimVolumeSource] = None
    secret: Optional[SecretVolumeSource] = None

    def create(self) -> Optional[Volume]:
        """Creates the Volume resource"""

        volume_name = self.volume_name
        logger.debug(f"Init Volume resource: {volume_name}")

        volume = Volume(name=volume_name)

        if self.volume_type == VolumeType.EMPTY_DIR:
            if self.empty_dir is not None and isinstance(
                self.empty_dir, EmptyDirVolumeSource
            ):
                volume.empty_dir = self.empty_dir
            else:
                volume.empty_dir = EmptyDirVolumeSource()
        elif self.volume_type == VolumeType.AWS_EBS:
            if self.aws_ebs is not None and isinstance(
                self.aws_ebs, AwsElasticBlockStoreVolumeSource
            ):
                volume.aws_elastic_block_store = self.aws_ebs
            else:
                print_error(
                    f"Volume {self.volume_type.value} selected but AwsElasticBlockStoreVolumeSource not provided."
                )
        elif self.volume_type == VolumeType.PERSISTENT_VOLUME_CLAIM:
            if self.pvc is not None and isinstance(
                self.pvc, PersistentVolumeClaimVolumeSource
            ):
                volume.persistent_volume_claim = self.pvc
            else:
                print_error(
                    f"Volume {self.volume_type.value} selected but PersistentVolumeClaimVolumeSource not provided."
                )
        elif self.volume_type == VolumeType.CONFIG_MAP:
            if self.config_map is not None and isinstance(
                self.config_map, ConfigMapVolumeSource
            ):
                volume.config_map = self.config_map
            else:
                print_error(
                    f"Volume {self.volume_type.value} selected but ConfigMapVolumeSource not provided."
                )
        elif self.volume_type == VolumeType.SECRET:
            if self.secret is not None and isinstance(self.secret, SecretVolumeSource):
                volume.secret = self.secret
            else:
                print_error(
                    f"Volume {self.volume_type.value} selected but SecretVolumeSource not provided."
                )
        elif self.volume_type == VolumeType.GCE_PERSISTENT_DISK:
            if self.gce_persistent_disk is not None and isinstance(
                self.gce_persistent_disk, GcePersistentDiskVolumeSource
            ):
                volume.gce_persistent_disk = self.gce_persistent_disk
            else:
                print_error(
                    f"Volume {self.volume_type.value} selected but GcePersistentDiskVolumeSource not provided."
                )
        elif self.volume_type == VolumeType.GIT_REPO:
            if self.git_repo is not None and isinstance(
                self.git_repo, GitRepoVolumeSource
            ):
                volume.git_repo = self.git_repo
            else:
                print_error(
                    f"Volume {self.volume_type.value} selected but GitRepoVolumeSource not provided."
                )
        elif self.volume_type == VolumeType.HOST_PATH:
            if self.host_path is not None and isinstance(
                self.host_path, HostPathVolumeSource
            ):
                volume.host_path = self.host_path
            else:
                print_error(
                    f"Volume {self.volume_type.value} selected but HostPathVolumeSource not provided."
                )

        return volume



from typing import Optional

from phidata.infra.aws.resource.group import AwsResourceGroup
from phidata.infra.base import InfraArgs
from phidata.infra.k8s.args import K8sArgs


class AwsArgs(InfraArgs):
    resources: Optional[AwsResourceGroup] = None
    resources_dir: str = "aws"

    # AWS configuration for this env
    # Use aws config from WorkspaceConfig if not provided
    aws_region: Optional[str] = None
    aws_profile: Optional[str] = None
    aws_config_file: Optional[str] = None
    aws_shared_credentials_file: Optional[str] = None



from typing import Any, Dict, Optional

from kubernetes.client.models.v1_object_meta import V1ObjectMeta
from pydantic import BaseModel, Field


class ObjectMeta(BaseModel):
    """
    ObjectMeta is metadata that all persisted resources must have,
    which includes all objects users must create.

    https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#objectmeta-v1-meta
    """

    resource_type: str = "ObjectMeta"

    # Name must be unique within a namespace. Is required when creating resources,
    # although some resources may allow a client to request the generation of an appropriate name automatically.
    # Name is primarily intended for creation idempotence and configuration definition.
    # Cannot be updated. More info: http://kubernetes.io/docs/user-guide/identifiers#names
    name: str
    # Namespace defines the space within which each name must be unique.
    # An empty namespace is equivalent to the "default" namespace, but "default" is the canonical representation.
    # Not all objects are required to be scoped to a namespace -
    # the value of this field for those objects will be empty. Must be a DNS_LABEL.
    # Cannot be updated. More info: http://kubernetes.io/docs/user-guide/namespaces
    namespace: Optional[str] = None
    # Map of string keys and values that can be used to organize and categorize (scope and select) objects.
    # May match selectors of replication controllers and services.
    # More info: http://kubernetes.io/docs/user-guide/labels
    labels: Optional[Dict[str, str]] = None
    # Annotations is an unstructured key value map stored with a resource that may be set by external tools
    # to store and retrieve arbitrary metadata. They are not queryable and should be preserved when
    # modifying objects. More info: http://kubernetes.io/docs/user-guide/annotations
    annotations: Optional[Dict[str, str]] = None
    # The name of the cluster which the object belongs to. This is used to distinguish resources with same name
    # and namespace in different clusters. This field is not set anywhere right now and apiserver is going
    # to ignore it if set in create or update request.
    cluster_name: Optional[str] = Field(None, alias="clusterName")

    def get_k8s_object(self) -> V1ObjectMeta:
        # Return a V1ObjectMeta object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_object_meta.py
        _v1_object_meta = V1ObjectMeta(
            name=self.name,
            namespace=self.namespace,
            labels=self.labels,
            annotations=self.annotations,
            cluster_name=self.cluster_name,
        )
        return _v1_object_meta

    class Config:
        arbitrary_types_allowed = True



from typing import Dict, List, Optional

from kubernetes.client import StorageV1Api
from kubernetes.client.models.v1_status import V1Status
from kubernetes.client.models.v1_storage_class import V1StorageClass
from kubernetes.client.models.v1_storage_class_list import V1StorageClassList
from pydantic import Field

from phidata.infra.k8s.api_client import K8sApiClient
from phidata.infra.k8s.resource.base import K8sResource
from phidata.utils.cli_console import print_info
from phidata.utils.log import logger


class StorageClass(K8sResource):
    """
    References:
        * Doc: https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#storageclass-v1-storage-k8s-io
        * Type: https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_storage_class.py
    """

    resource_type: str = "StorageClass"

    # AllowVolumeExpansion shows whether the storage class allow volume expand
    allow_volume_expansion: Optional[str] = Field(None, alias="allowVolumeExpansion")
    # Dynamically provisioned PersistentVolumes of this storage class are created with these mountOptions,
    # e.g. ["ro", "soft"]. Not validated - mount of the PVs will simply fail if one is invalid.
    mount_options: Optional[List[str]] = Field(None, alias="mountOptions")
    # Parameters holds the parameters for the provisioner that should create volumes of this storage class.
    parameters: Dict[str, str]
    # Provisioner indicates the type of the provisioner.
    provisioner: str
    # Dynamically provisioned PersistentVolumes of this storage class are created with this reclaimPolicy.
    # Defaults to Delete.
    reclaim_policy: Optional[str] = Field(None, alias="reclaimPolicy")
    # VolumeBindingMode indicates how PersistentVolumeClaims should be provisioned and bound.
    # When unset, VolumeBindingImmediate is used.
    # This field is only honored by servers that enable the VolumeScheduling feature.
    volume_binding_mode: Optional[str] = Field(None, alias="volumeBindingMode")

    # List of attributes to include in the K8s manifest
    fields_for_k8s_manifest: List[str] = [
        "allow_volume_expansion",
        "mount_options",
        "parameters",
        "provisioner",
        "reclaim_policy",
        "volume_binding_mode",
    ]

    def get_k8s_object(self) -> V1StorageClass:
        """Creates a body for this StorageClass"""

        # Return a V1StorageClass object to create a StorageClass
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_storage_class.py
        _v1_storage_class = V1StorageClass(
            allow_volume_expansion=self.allow_volume_expansion,
            api_version=self.api_version,
            kind=self.kind,
            metadata=self.metadata.get_k8s_object(),
            mount_options=self.mount_options,
            provisioner=self.provisioner,
            parameters=self.parameters,
            reclaim_policy=self.reclaim_policy,
            volume_binding_mode=self.volume_binding_mode,
        )
        return _v1_storage_class

    @staticmethod
    def get_from_cluster(
        k8s_client: K8sApiClient, namespace: Optional[str] = None, **kwargs
    ) -> Optional[List[V1StorageClass]]:
        """Reads StorageClasses from K8s cluster.

        Args:
            k8s_client: K8sApiClient for the cluster
            namespace: Namespace to use.
        """
        storage_v1_api: StorageV1Api = k8s_client.storage_v1_api
        sc_list: Optional[V1StorageClassList] = storage_v1_api.list_storage_class()
        scs: Optional[List[V1StorageClass]] = None
        if sc_list:
            scs = sc_list.items
            # logger.debug(f"scs: {scs}")
            # logger.debug(f"scs type: {type(scs)}")
        return scs

    def _create(self, k8s_client: K8sApiClient) -> bool:

        storage_v1_api: StorageV1Api = k8s_client.storage_v1_api
        k8s_object: V1StorageClass = self.get_k8s_object()

        logger.debug("Creating: {}".format(self.get_resource_name()))
        v1_storage_class: V1StorageClass = storage_v1_api.create_storage_class(
            body=k8s_object,
            async_req=self.async_req,
            pretty=self.pretty,
        )
        # logger.debug("Created: {}".format(v1_storage_class))
        if v1_storage_class.metadata.creation_timestamp is not None:
            logger.debug("StorageClass Created")
            self.active_resource = v1_storage_class
            self.active_resource_class = V1StorageClass
            return True
        logger.error("StorageClass could not be created")
        return False

    def _read(self, k8s_client: K8sApiClient) -> Optional[V1StorageClass]:
        """Returns the "Active" StorageClass from the cluster"""

        active_resource: Optional[V1StorageClass] = None
        active_resources: Optional[List[V1StorageClass]] = self.get_from_cluster(
            k8s_client=k8s_client,
        )
        # logger.debug(f"Active Resources: {active_resources}")
        if active_resources is None:
            return None

        active_resources_dict = {_sc.metadata.name: _sc for _sc in active_resources}

        sc_name = self.get_resource_name()
        if sc_name in active_resources_dict:
            active_resource = active_resources_dict[sc_name]
            self.active_resource = active_resource
            self.active_resource_class = V1StorageClass
            logger.debug(f"Found active {sc_name}")
        return active_resource

    def _update(self, k8s_client: K8sApiClient) -> bool:

        storage_v1_api: StorageV1Api = k8s_client.storage_v1_api
        sc_name = self.get_resource_name()
        k8s_object: V1StorageClass = self.get_k8s_object()

        logger.debug("Updating: {}".format(sc_name))
        v1_storage_class: V1StorageClass = storage_v1_api.patch_storage_class(
            name=sc_name,
            body=k8s_object,
            async_req=self.async_req,
            pretty=self.pretty,
        )
        # logger.debug("Updated:\n{}".format(pformat(v1_storage_class.to_dict(), indent=2)))
        if v1_storage_class.metadata.creation_timestamp is not None:
            logger.debug("StorageClass Updated")
            self.active_resource = v1_storage_class
            self.active_resource_class = V1StorageClass
            return True
        logger.error("StorageClass could not be updated")
        return False

    def _delete(self, k8s_client: K8sApiClient) -> bool:

        storage_v1_api: StorageV1Api = k8s_client.storage_v1_api
        sc_name = self.get_resource_name()

        logger.debug("Deleting: {}".format(sc_name))
        self.active_resource = None
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_status.py
        delete_status: V1Status = storage_v1_api.delete_storage_class(
            name=sc_name,
            async_req=self.async_req,
            pretty=self.pretty,
        )
        logger.debug("delete_status: {}".format(delete_status.status))
        if delete_status.status == "Success":
            logger.debug("StorageClass Deleted")
            return True
        logger.error("StorageClass could not be deleted")
        return False



from typing import Optional, List, Dict
from typing_extensions import Literal

from phidata.infra.base import InfraConfig, InfraArgs
from phidata.infra.docker.config import DockerConfig
from phidata.infra.k8s.config import K8sConfig
from phidata.infra.aws.config import AwsConfig
from phidata.utils.cli_console import print_error


class WorkspaceConfigArgs(InfraArgs):
    # default env for phi ws up/down
    default_env: Optional[Literal["dev", "stg", "prd"]] = None
    # default config for phi ws up/down
    default_config: Optional[str] = None

    docker: Optional[List[DockerConfig]] = None
    k8s: Optional[List[K8sConfig]] = None
    # AWS configuration per env
    aws: Optional[List[AwsConfig]] = None
    # Common aws configuration used by resources and data assets
    # this is passed down to each aws config and set in the local_env
    aws_region: Optional[str] = None
    aws_profile: Optional[str] = None
    aws_config_file: Optional[str] = None
    aws_shared_credentials_file: Optional[str] = None

    # Path to important directories relative to the workspace_root
    # These directories are joined to the workspace_root dir
    #   to build paths depending on the environments (local, docker, k8s)
    # defaults are set by WorkspaceConfig.__init__()
    scripts_dir: str
    storage_dir: str
    meta_dir: str
    products_dir: str
    notebooks_dir: str
    workspace_config_dir: str


class WorkspaceConfig(InfraConfig):
    def __init__(
        self,
        version: Optional[str] = None,
        enabled: bool = True,
        # default env for phi ws up/down
        default_env: Optional[Literal["dev", "stg", "prd"]] = None,
        # default config for phi ws up/down
        default_config: Optional[str] = None,
        docker: Optional[List[DockerConfig]] = None,
        k8s: Optional[List[K8sConfig]] = None,
        # AWS configuration per env
        aws: Optional[List[AwsConfig]] = None,
        # Common aws configuration used by resources and data assets
        # this is passed down to each aws config and set in the local_env
        aws_region: Optional[str] = None,
        aws_profile: Optional[str] = None,
        aws_config_file: Optional[str] = None,
        aws_shared_credentials_file: Optional[str] = None,
        # src_dir is the dir containing the products directory.
        src_dir: str = "data",
        # path to important directories relative to the workspace_root
        scripts_dir: Optional[str] = None,
        storage_dir: Optional[str] = None,
        meta_dir: Optional[str] = None,
        # path to important directories relative to the
        # workspace_root/src_dir
        products_dir: Optional[str] = None,
        notebooks_dir: Optional[str] = None,
        workspace_config_dir: Optional[str] = None,
        # Env vars added to local env when running workflows
        local_env: Optional[Dict[str, str]] = None,
        # Env vars added to docker env when running workflows
        docker_env: Optional[Dict[str, str]] = None,
        # Env vars added to k8s env when running workflows
        k8s_env: Optional[Dict[str, str]] = None,
    ):
        super().__init__()
        scripts_dir_clean = scripts_dir if scripts_dir else f"scripts"
        storage_dir_clean = storage_dir if storage_dir else f"storage"
        meta_dir_clean = meta_dir if meta_dir else f"meta"
        products_dir_clean = products_dir if products_dir else f"{src_dir}/products"
        notebooks_dir_clean = notebooks_dir if notebooks_dir else f"{src_dir}/notebooks"
        workspace_config_dir_clean = (
            workspace_config_dir if workspace_config_dir else f"{src_dir}/workspace"
        )

        try:
            self.args: WorkspaceConfigArgs = WorkspaceConfigArgs(
                version=version,
                enabled=enabled,
                default_env=default_env,
                default_config=default_config,
                docker=docker,
                k8s=k8s,
                aws=aws,
                aws_region=aws_region,
                aws_profile=aws_profile,
                aws_config_file=aws_config_file,
                aws_shared_credentials_file=aws_shared_credentials_file,
                scripts_dir=scripts_dir_clean,
                storage_dir=storage_dir_clean,
                meta_dir=meta_dir_clean,
                products_dir=products_dir_clean,
                notebooks_dir=notebooks_dir_clean,
                workspace_config_dir=workspace_config_dir_clean,
                local_env=local_env,
                docker_env=docker_env,
                k8s_env=k8s_env,
            )
        except Exception as e:
            print_error(f"Args for {self.__class__.__name__} are not valid")
            raise

    @property
    def default_env(self) -> Optional[Literal["dev", "stg", "prd"]]:
        return self.args.default_env if self.args else None

    @property
    def default_config(self) -> Optional[str]:
        return self.args.default_config if self.args else None

    @property
    def docker(self) -> Optional[List[DockerConfig]]:
        return self.args.docker if self.args else None

    @property
    def k8s(self) -> Optional[List[K8sConfig]]:
        return self.args.k8s if self.args else None

    @property
    def aws(self) -> Optional[List[AwsConfig]]:
        return self.args.aws if self.args else None

    @property
    def aws_region(self) -> Optional[str]:
        return self.args.aws_region if self.args else None

    @property
    def aws_profile(self) -> Optional[str]:
        return self.args.aws_profile if self.args else None

    @property
    def aws_config_file(self) -> Optional[str]:
        return self.args.aws_config_file if self.args else None

    @property
    def aws_shared_credentials_file(self) -> Optional[str]:
        return self.args.aws_shared_credentials_file if self.args else None

    def is_valid(self) -> bool:
        if self.docker is not None:
            if not isinstance(self.docker, list):
                raise ValueError("docker should be a list")
            for dc in self.docker:
                if not isinstance(dc, DockerConfig):
                    raise ValueError(f"Invalid DockerConfig: {dc}")

        if self.k8s is not None:
            if not isinstance(self.k8s, list):
                raise ValueError("k8s should be a list")
            for kc in self.k8s:
                if not isinstance(kc, K8sConfig):
                    raise ValueError(f"Invalid K8sConfig: {kc}")

        if self.aws is not None:
            if not isinstance(self.aws, list):
                raise ValueError("aws should be a list")
            for awsc in self.aws:
                if not isinstance(awsc, AwsConfig):
                    raise ValueError(f"Invalid AwsConfig: {awsc}")
        return True



from typing import List, Optional, Dict

from pydantic import BaseModel

from phidata.infra.k8s.create.common.port import CreatePort
from phidata.infra.k8s.create.core.v1.volume import CreateVolume
from phidata.infra.k8s.enums.image_pull_policy import ImagePullPolicy
from phidata.utils.common import get_image_str
from phidata.infra.k8s.resource.core.v1.container import (
    Container,
    ContainerPort,
    EnvFromSource,
    VolumeMount,
    ConfigMapEnvSource,
    SecretEnvSource,
    EnvVar,
    EnvVarSource,
    ConfigMapKeySelector,
    SecretKeySelector,
)
from phidata.utils.log import logger


class CreateEnvVarFromConfigMap(BaseModel):
    env_var_name: str
    configmap_name: str
    configmap_key: Optional[str] = None


class CreateEnvVarFromSecret(BaseModel):
    env_var_name: str
    secret_name: str
    secret_key: Optional[str] = None


class CreateContainer(BaseModel):
    container_name: str
    app_name: str
    image_name: str
    image_tag: str
    args: Optional[List[str]] = None
    command: Optional[List[str]] = None
    image_pull_policy: Optional[ImagePullPolicy] = ImagePullPolicy.IF_NOT_PRESENT
    env: Optional[Dict[str, str]] = None
    envs_from_configmap: Optional[List[str]] = None
    envs_from_secret: Optional[List[str]] = None
    env_vars_from_secret: Optional[List[CreateEnvVarFromSecret]] = None
    env_vars_from_configmap: Optional[List[CreateEnvVarFromConfigMap]] = None
    ports: Optional[List[CreatePort]] = None
    volumes: Optional[List[CreateVolume]] = None
    labels: Optional[Dict[str, str]] = None

    def create(self) -> Optional[Container]:
        """Creates the Container resource"""

        container_name = self.container_name
        logger.debug(f"Init Container resource: {container_name}")

        container_ports: Optional[List[ContainerPort]] = None
        if self.ports:
            container_ports = []
            for _port in self.ports:
                container_ports.append(
                    ContainerPort(
                        name=_port.name,
                        container_port=_port.container_port,
                        protocol=_port.protocol,
                    )
                )

        env_from: Optional[List[EnvFromSource]] = None
        if self.envs_from_configmap:
            if env_from is None:
                env_from = []
            for _cm_name_for_env in self.envs_from_configmap:
                env_from.append(
                    EnvFromSource(
                        config_map_ref=ConfigMapEnvSource(name=_cm_name_for_env)
                    )
                )
        if self.envs_from_secret:
            if env_from is None:
                env_from = []
            for _secretenvs in self.envs_from_secret:
                env_from.append(
                    EnvFromSource(secret_ref=SecretEnvSource(name=_secretenvs))
                )

        env: Optional[List[EnvVar]] = None
        if self.env is not None and isinstance(self.env, dict):
            if env is None:
                env = []
            for key, value in self.env.items():
                env.append(
                    EnvVar(
                        name=key,
                        value=value,
                    )
                )

        if self.env_vars_from_configmap:
            if env is None:
                env = []
            for _cmenv_var in self.env_vars_from_configmap:
                env.append(
                    EnvVar(
                        name=_cmenv_var.env_var_name,
                        value_from=EnvVarSource(
                            config_map_key_ref=ConfigMapKeySelector(
                                key=_cmenv_var.configmap_key
                                if _cmenv_var.configmap_key
                                else _cmenv_var.env_var_name,
                                name=_cmenv_var.configmap_name,
                            )
                        ),
                    )
                )
        if self.env_vars_from_secret:
            if env is None:
                env = []
            for _secretenv_var in self.env_vars_from_secret:
                env.append(
                    EnvVar(
                        name=_secretenv_var.env_var_name,
                        value_from=EnvVarSource(
                            secret_key_ref=SecretKeySelector(
                                key=_secretenv_var.secret_key
                                if _secretenv_var.secret_key
                                else _secretenv_var.env_var_name,
                                name=_secretenv_var.secret_name,
                            )
                        ),
                    )
                )

        volume_mounts: Optional[List[VolumeMount]] = None
        if self.volumes:
            volume_mounts = []
            for _volume in self.volumes:
                volume_mounts.append(
                    VolumeMount(
                        name=_volume.volume_name,
                        mount_path=_volume.mount_path,
                    )
                )

        container_resource = Container(
            name=container_name,
            image=get_image_str(self.image_name, self.image_tag),
            image_pull_policy=self.image_pull_policy,
            args=self.args,
            command=self.command,
            ports=container_ports,
            env_from=env_from,
            env=env,
            volume_mounts=volume_mounts,
        )
        return container_resource
